
/*
 *  BATCH.C
 *
 *  (1) Process incomming file(s) bms_in, call receiver
 *  (2) Process Comm files with their alert flag set, modify com_TxOk
 *  (3) Call Comm files with their alert flag set
 *  (4) Call Comm files that have timed out
 */

#include "defs.h"
#include "config.h"
#include "lib_protos.h"

Prototype long TotalBytesSent;
Prototype long TotalBytesRecv;
Prototype long MailBytes;
Prototype long MinMailBytes;

Prototype  void BatchBMS(void);
Prototype  void BatchBMSReceive(FILE *);
Prototype  void BatchBMSTransmit(void);
Prototype  void BMSReceiveScanDatabase(DBHan *);
Prototype record_t WriteBMSChannelRecord(DBHan *, record_t, record_t, Comm *);

long TotalBytesSent;
long TotalBytesRecv;

long MailBytes;
long MinMailBytes;

void
BatchBMS()
{
    /*
     *	access bms_in mail box
     */

    {
	char *mailbox;
	char *mailtmp;
	FILE *fi;


	mailbox = zstrdup(MakeConfigPath(UUMAIL, "bms_in"));
	mailtmp = zstrdup(MakeConfigPath(UUMAIL, "bms_in.old"));
	LockFile(mailbox);
	if (fi = fopen(mailbox, "r")) {
	    BatchBMSReceive(fi);
	    fclose(fi);
	    remove(mailtmp);
	    rename(mailbox, mailtmp);
	}
	UnLockFile(mailbox);
    }

    BatchBMSTransmit();
}

void
BatchBMSTransmit()
{
    /*
     *	Ok, here's the problem... we have a certain amount of data pending
     *	for transmission and need to determine how much of it we can send.
     *
     *	(1) Calculate the maximum amount of data sendable for each Comm
     *	    node based on the number of bytes ready to send, the maximum
     *	    the remote is willing to receive per day, and cap it to what
     *	    bandwidth we have remaining (daily, weekly)
     *
     *	    The result is a per node maximum that does not take into
     *	    account other nodes that may have data to send as well.
     *
     *	(2) Total the results from (1) together and determine how much over
     *	    our daily/weekly limit we are for the worst case (lowest limit).
     *	    figure the percentage of the total that we can handle.
     */

    {
	RegNode *rn;
	CtlNode *cn;

	for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	    long tx24 = (rn->rn_Param.pm_MaxTxBpd) ? rn->rn_Param.pm_MaxTxBpd : DefRN->rn_Param.pm_MaxTxBpd;
	    long txweek = (rn->rn_Param.pm_MaxTxBpw) ? rn->rn_Param.pm_MaxTxBpw : DefRN->rn_Param.pm_MaxTxBpw;

	    LoadBMSRegTree(rn);

	    rn->rn_Comm.com_TxOk = 100000;  /*	registration comm always allowed    */

	    if (rn->rn_Comm.com_DID == 0)   /*  don't run queued cmds if incomplete registration */
		continue;

	    for (cn = GetHead(&rn->rn_CtlList); cn; cn = GetSucc(&cn->cn_Node)) {
		cn->cn_Comm.com_TxOk = 0;
		cn->cn_Comm.com_TmpTx = cn->cn_Comm.com_TxReq;
		if (cn->cn_Comm.com_TmpTx > cn->cn_Comm.com_RemTxBpd)
		    cn->cn_Comm.com_TmpTx = cn->cn_Comm.com_RemTxBpd;
		if (cn->cn_Comm.com_TmpTx > cn->cn_Comm.com_RemTxBpw)
		    cn->cn_Comm.com_TmpTx = cn->cn_Comm.com_RemTxBpw;
		if (cn->cn_Comm.com_TmpTx > tx24)
		    cn->cn_Comm.com_TmpTx = tx24;
		if (cn->cn_Comm.com_TmpTx > txweek)
		    cn->cn_Comm.com_TmpTx = txweek;
	    }
	}
    }

    /*
     *	(3) Loop over all data starting with the lowest allowed
     *	    daily/weekly limits and looping until the highest allowed
     *	    daily/weekly limits.  For each looping we only apply the
     *	    result to the nodes which >= the chosen daily/weekly limits
     *
     *	    E.G. we apply the first round to ALL nodes, the second round
     *	    to fewer (higher limit) nodes, the next round, etc...
     *
     *	    Each round calculates the total transmittable bytes
     *	    (com_TmpTx) verses our limits for that round,
     *	    for those nodes that meet the requirements.
     *
     *	    We then compute the percentage of transmittable bytes so
     *	    as not to go over our limits for the round and apply it
     *	    to all applicable modes, subtracting from com_TmpTx and
     *	    adding to com_TxOk.
     */

    {
	long	mindaily = 0;
	long	minweekly = 0;
	long	used_daily;
	long	used_weekly;
	short	round = 0;
	RegNode *rn;

	/*
	 *  get total usage for the day and the week.  Make this our
	 *  minimum baseline
	 */

	for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node))
	    CalculateHistogramSent(rn, &mindaily, &minweekly);

	used_daily = mindaily;
	used_weekly = minweekly;

	for (;;) {
	    long    maxdaily = 0x7FFFFFFF;
	    long    maxweekly = 0x7FFFFFFF;
	    long    want_out;
	    long    max_out;
	    long    perc;

	    /*
	     *	calculate next round
	     */

	    ++round;

	    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
		{
		    long dayLimit = rn->rn_Param.pm_MaxTxBpd;
		    if (dayLimit == 0)
			dayLimit = DefRN->rn_Param.pm_MaxTxBpd;
		    if (dayLimit > mindaily && dayLimit < maxdaily)
			maxdaily = dayLimit;
		}
		{
		    long wkLimit = rn->rn_Param.pm_MaxTxBpw;
		    if (wkLimit == 0)
			wkLimit = DefRN->rn_Param.pm_MaxTxBpw;
		    if (wkLimit > minweekly && wkLimit < maxweekly)
			maxweekly = wkLimit;
		}
	    }

	    /*
	     *	We now have maxweekly and maxdaily for this round.
	     *	We apply it to all nodes with >= maxweekly and maxdaily
	     *
	     *	(1) determine maximum number of bytes outstanding.  We count
	     *	only the difference between the last run and this run else
	     *	we would be counting previous runs multiple times!
	     */

	    if (maxdaily == 0x7FFFFFFF)
		max_out = 0;
	    else
		max_out = maxdaily - mindaily;

	    if (maxweekly != 0x7FFFFFFF && max_out > maxweekly - minweekly)
		max_out = maxweekly - minweekly;

	    if (max_out < 0)
		max_out = 0;

	    want_out = 0;

	    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
		if (rn->rn_Comm.com_DID == 0)
		    continue;

		/*
		 *  Skip nodes that do not apply to this round.  All
		 *  nodes apply to the first round even if they are
		 *  less the maxdaily due to data already sent, so
		 *  we report the right numbers!
		 */

		{
		    long dayLimit = rn->rn_Param.pm_MaxTxBpd;
		    if (dayLimit == 0)
			dayLimit = DefRN->rn_Param.pm_MaxTxBpd;
		    if (dayLimit > used_daily && dayLimit < maxdaily)
			continue;
		}
		{
		    long wkLimit = rn->rn_Param.pm_MaxTxBpw;
		    if (wkLimit == 0)
			wkLimit = DefRN->rn_Param.pm_MaxTxBpw;
		    if (wkLimit > used_weekly && wkLimit < maxweekly)
			continue;
		}
		{
		    CtlNode *cn;

		    for (cn = GetHead(&rn->rn_CtlList); cn; cn = GetSucc(&cn->cn_Node)) {
			want_out += cn->cn_Comm.com_TmpTx;
		    }
		}
	    }

	    /*
	     *	We now have want_out and max_out, calculate the percentage
	     *	we can take to ensure that output is not larger then
	     *	max_out.
	     *
	     *	want_out * N / 1000 = max_out
	     *
	     *	    max_out * 1000
	     *	N = ---------------	    (N in 1/10000 units)
	     *		want_out
	     */

	    perc = (want_out) ? max_out * 1000 / want_out : 1000;

	    bmslog("Batchout", "Round #%d  %d/%d bytes batchout, %d.%01d%% of reqs handled",
		round,
		max_out,
		want_out,
		perc / 10,
		perc % 10
	    );

	    if (max_out == 0)
		break;

	    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
		if (rn->rn_Comm.com_DID == 0)   /*  don't run queued cmds if incomplete registration */
		    continue;
		{
		    long dayLimit = rn->rn_Param.pm_MaxTxBpd;
		    if (dayLimit == 0)
			dayLimit = DefRN->rn_Param.pm_MaxTxBpd;
		    if (dayLimit > used_daily && dayLimit < maxdaily)
			continue;
		}
		{
		    long wkLimit = rn->rn_Param.pm_MaxTxBpw;
		    if (wkLimit == 0)
			wkLimit = DefRN->rn_Param.pm_MaxTxBpw;
		    if (wkLimit > used_weekly && wkLimit < maxweekly)
			continue;
		}
		{
		    CtlNode *cn;

		    for (cn = GetHead(&rn->rn_CtlList); cn; cn = GetSucc(&cn->cn_Node)) {
			if (perc > 1000) {
			    /*
			     *	greater then 100%
			     */

			    cn->cn_Comm.com_TxOk += cn->cn_Comm.com_TmpTx;
			    cn->cn_Comm.com_TmpTx = 0;
			} else {
			    /*
			     *	less then 100%
			     */

			    long n = perc * cn->cn_Comm.com_TmpTx / 1000;
			    cn->cn_Comm.com_TxOk += n;
			    cn->cn_Comm.com_TmpTx -= n;
			}
			if (cn->cn_Comm.com_TxOk > cn->cn_Comm.com_TxReq)
			    cn->cn_Comm.com_TxOk = cn->cn_Comm.com_TxReq;
			if (cn->cn_Comm.com_TxOk < 0)
			    cn->cn_Comm.com_TxOk = 0;
		    }
		}
	    }

	    /*
	     *	Setup for next loop
	     */

	    mindaily = maxdaily;
	    minweekly = maxweekly;
	}
    }

    /*
     *	(4) Call all nodes with transmit data pending or that are alerted,
     *	    then save and unload.
     *
     *	    We create a database for each potential node and allow
     *	    the dispatched routine to fill it.	This database, if
     *	    no longer empty, is then email'd and deleted
     *
     *	    Data is further limited by the MinMail variable
     */

    for (;;) {
	RegNode *rn;
	CtlNode *cn;
	char *tmpFileName = AllocTemporaryDiskFile();
	DBHan *db = NULL;
	short done = 1;

	if (DebugOpt)
	    bmslog("Batchout", "Top");

	for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	    long size;
	    long type;

	    /*
	     *	Get clean database file (if clean from last run then
	     *	we don't have to close/reopen)
	     */

	    if (db == NULL)
		db = OpenDataBase(tmpFileName, "w");
	    if (db == NULL) {
		bmslog("Batch", "database reopen failed: %s", tmpFileName);
		break;
	    }

	    /*
	     *	Calculate mail parameters
	     */

	    MailBytes = (rn->rn_Param.pm_MaxTxMail) ? rn->rn_Param.pm_MaxTxMail : DefRN->rn_Param.pm_MaxTxMail;
	    MinMailBytes = (rn->rn_Param.pm_MinMail) ? rn->rn_Param.pm_MinMail : DefRN->rn_Param.pm_MinMail;

	    if (rn->rn_Comm.com_Flags & GF_ALERT) {
		rn->rn_Comm.com_Flags &= ~GF_ALERT;
		rn->rn_Comm.com_Flags |= GF_MODIFIED;
		DispatchCommTx(db, 0, rn, NULL, &rn->rn_Comm);
	    }
	    if (rn->rn_Comm.com_TOTime >= 0) {
		long n = rn->rn_Comm.com_TOStamp - gettime();

		if (n > rn->rn_Comm.com_TOTime)
		    n = rn->rn_Comm.com_TOTime;
		if (n > 0 && n > rn->rn_Comm.com_TOTime - 60*60)
		    n = rn->rn_Comm.com_TOTime;
		if (rn->rn_Comm.com_TOTime != n) {
		    rn->rn_Comm.com_TOTime = n;
		    rn->rn_Flags |= GF_MODIFIED;
		}
		if (n < 0) {
		    DispatchCommTo(db, 0, rn, NULL, &rn->rn_Comm);
		}
	    }

	    /*
	     *	Do not run the state machines for control nodes of a node
	     *	whos registration is still in progress.
	     */

	    if (rn->rn_Comm.com_DID) {
		for (cn = GetHead(&rn->rn_CtlList); cn; cn = GetSucc(&cn->cn_Node)) {
		    /*
		     *	check transmitter
		     */

		    if (cn->cn_Comm.com_TxOk || (cn->cn_Comm.com_Flags & GF_ALERT)) {
			cn->cn_Comm.com_Flags &= ~GF_ALERT;
			cn->cn_Comm.com_Flags |= GF_MODIFIED;
			DispatchCommTx(db, 0, rn, cn, &cn->cn_Comm);
		    }

		    /*
		     *	check for a timeout.  n < 0 indicates a timeout,
		     *	n > cn->cn_Comm.com_TOTime indicates an error
		     *	(somebody reset the time).  cn->cn_Comm.com_TOTime < 0
		     *	indicates the timeout is disabled.
		     *
		     *	TOTime acts as a double check and need not be exact,
		     *	so to prevent undo writing to the database we do not
		     *	modify it if its within an hour.
		     */

		    if (cn->cn_Comm.com_TOTime >= 0) {
			long n = cn->cn_Comm.com_TOStamp - gettime();

			if (n > cn->cn_Comm.com_TOTime)
			    n = cn->cn_Comm.com_TOTime;
			if (n > 0 && n > cn->cn_Comm.com_TOTime - 60*60)
			    n = cn->cn_Comm.com_TOTime;
			if (cn->cn_Comm.com_TOTime != n) {
			    cn->cn_Comm.com_TOTime = n;
			    cn->cn_Flags |= GF_MODIFIED;
			}
			if (n < 0) {
			    DispatchCommTo(db, 0, rn, cn, &cn->cn_Comm);
			}
		    }

		    /*
		     *	if MailBytes <= MinMailBytes then close/reopen to
		     *	split mail
		     */

		    if (MailBytes <= MinMailBytes) {
			MailBytes = (rn->rn_Param.pm_MaxTxMail) ? rn->rn_Param.pm_MaxTxMail : DefRN->rn_Param.pm_MaxTxMail;
			CloseDataBase(db);
			MailDataBase(rn, tmpFileName);
			done = 0;
			bmslog("Batchout", "Mail sent to %s", rn->rn_Node.ln_Name);
			db = OpenDataBase(tmpFileName, "w");
			if (db == NULL) {
			    bmslog("Batch", "database reopen failed: %s", tmpFileName);
			    break;
			}
		    }
		}
	    }
	    if (db) {
		if (ScanRecord(db, 0, &size, &type)) {
		    CloseDataBase(db);
		    MailDataBase(rn, tmpFileName);
		    done = 0;
		    bmslog("Batchout", "Mail sent to %s", rn->rn_Node.ln_Name);
		    db = NULL;
		}
	    }
	    SaveBMSDataBase();
	}
	if (db) {
	    CloseDataBase(db);
	    db = NULL;
	}
	remove(tmpFileName);
	zfree(tmpFileName);

	if (done)
	    break;
    }
    {
	RegNode *rn;

	SaveBMSDataBase();
	for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node))
	    UnLoadBMSRegTree(rn);
    }
}

/*
 *  BATCHBMSRECEIVE()
 *
 *  We have received a BMS request.  BMS mail is always in the form of a
 *  database.  For each BMS encoded block we unpack it then scan it,
 *  dispatching the database/rid to the appropriate state machine.
 */


void
BatchBMSReceive(fi)
FILE *fi;
{
    char *tmpFileName = AllocTemporaryDiskFile();
    FILE *fo = NULL;

    for (;;) {
	if (fo = fopen(tmpFileName, "w")) {
	    DBHan *db;
	    long r;

	    if ((r = BMSDecodeFile(fi, fo)) < 0) {
		if (r == -1)
		    break;
	    }
	    fclose(fo);
	    fo = NULL;
	    if (r == 0) {
		if (db = OpenDataBase(tmpFileName, "r")) {
		    BMSReceiveScanDatabase(db);
		    CloseDataBase(db);
		}
	    }
	}
    }
    if (fo)
	fclose(fo);
    remove(tmpFileName);
    zfree(tmpFileName);
}

void
BMSReceiveScanDatabase(db)
DBHan *db;
{
    record_t rid;
    long type;
    long size;
    char *ptr;
    RegNode *rn = NULL;

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node))
	LoadBMSRegTree(rn);

    for (rid = ScanRecord(db, 0, &size, &type); rid; rid = ScanNextRecord(db, rid, &size, &type)) {
	switch(type) {
	case DBTYPE_BMS_REGISTER:
	    ptr = LoadRecord(db, rid, NULL, size + 1);
	    SMRxNoChannel(db, rid);
	    zfree(ptr);
	    break;
	case DBTYPE_BMS_CHANNEL:
	    {
		long loid;
		long rmid;
		char *tmp;

		ptr = LoadRecord(db, rid, NULL, size + 2);
		rmid = strtol(ptr, &tmp, 0);
		loid = strtol(tmp + 1, &tmp, 0);

		if (DebugOpt)
		    printf("RECEIVED CHANNEL PACKET: %s FROM %d TO %d\n", ptr, rmid, loid);

		/*
		 *  Find the comm node associated with the channel
		 */

		for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
		    CtlNode *cn;

		    if (rn->rn_Comm.com_SID == loid) {
			if (rn->rn_Comm.com_DID == 0)
			    rn->rn_Comm.com_DID = rmid;
			/*
			if (rn->rn_Comm.com_DID != rmid) {
			    bmslog("Recv", "Remote ID mismatch loid=%d rmid=%d", loid, rmid);
			    break;
			}
			*/
			DispatchCommRx(db, rid, rn, NULL, &rn->rn_Comm);
			break;
		    }
		    for (cn = GetHead(&rn->rn_CtlList); cn; cn = GetSucc(&cn->cn_Node)) {
			if (cn->cn_Comm.com_SID == loid) {
			    if (cn->cn_Comm.com_DID == 0)
				cn->cn_Comm.com_DID = rmid;
			    if (cn->cn_Comm.com_DID != rmid) {
				bmslog("Recv", "Remote ID mismatch channel loid=%d rmid=%d", loid, rmid);
				break;
			    }

			    /*
			     *	On reception of any data packet through a
			     *	valid registration, reset timeout to
			     *	6 months for the registration
			     */

			    SetCommTimeout(&rn->rn_Comm, ONEMONTH * 6);
			    DispatchCommRx(db, rid, rn, cn, &cn->cn_Comm);
			    break;
			}
		    }
		    if (cn)
			break;
		}
		if (rn == NULL)
		    bmslog("Recv", "Local ID, no channel: loid=%d rmid=%d\n", loid, rmid);
	    }
	    zfree(ptr);
	    break;
	default:
	    bmslog("Recv", "Unknown database ID: 0x%08lx size %d\n", type, size);
	    break;
	}
    }

    SaveBMSDataBase();
    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node))
	UnLoadBMSRegTree(rn);
}

record_t
WriteBMSChannelRecord(db, pid, rid, comm)
DBHan *db;
record_t pid;
record_t rid;
Comm *comm;
{
    char buf[64];

    sprintf(buf, "0x%lx:0x%lx", comm->com_SID, comm->com_DID);
    rid = WriteRecord(db, pid, rid, DBTYPE_BMS_CHANNEL, buf, strlen(buf));
    return(rid);
}

