
/*
 *  BMSDB.C
 *
 *  BMS database operations
 */

#include "defs.h"

Prototype void LoadBMSTopLevel(void);
Prototype void LoadBMSRegTree(RegNode *);
Prototype void LoadBMSCtlTree(CtlNode *);
Prototype void LoadBMSFileAccessTree(RegNode *, record_t);
Prototype void LoadBMSFileTree(FileNode *);
Prototype void LoadBMSChunkTree(FileNode *);
Prototype void LoadBMSCommTree(DBHan *, record_t, Comm *);

Prototype void SaveBMSDataBase(void);
Prototype void SaveBMSCtlList(List *, record_t);
Prototype void SaveBMSFileList(List *, record_t);
Prototype void SaveBMSChunkList(FileNode *);
Prototype void SaveBMSFileAccessTree(RegNode *);

Prototype void UnLoadBMSDataBase(void);
Prototype void UnLoadBMSRegTree(RegNode *);
Prototype void UnLoadBMSCtlTree(CtlNode *);
Prototype void UnLoadBMSFileTree(FileNode *);
Prototype void UnLoadBMSFileAccessTree(FileAccess *);

Prototype RegNode *AllocRegNode(void);
Prototype CtlNode *AllocCtlNode(RegNode *);
Prototype FileNode *AllocFileNode(CtlNode *);
Prototype FileChunk *AllocFileChunk(FileNode *, long, long);
Prototype FileAccess *AllocFileAccess(RegNode *, const char *);
Prototype void AddFileChunk(FileNode *, FileChunk *);

/*
 *  LoadBMSTopLevel(void)
 *
 *  Load top level registration nodes.
 *
 *  Keep in mind that when calling LoadRecord(), any space beyond the
 *  actual size of the database record is cleared.
 *
 *  Note that the redundant clear of GF_LOADED will prevent problems
 *  with future planned enhancements (greater incrementality of loading)
 */

void
LoadBMSTopLevel(void)
{
    record_t rid;
    long size;
    long type;

    Db = OpenDataBase("BMS:BMS.DB", "r");
    if (Db == NULL) {
	bmslog("LoadDB", "Cannot open BMS:BMS.DB");
	exit(20);
    }
    for (rid = ScanRecord(Db, 0, &size, &type); rid; rid = ScanNextRecord(Db, rid, &size, &type)) {
	if (type == DBTYPE_REG) {
	    RegNode *rn;

	    if (rn = LoadExtRecord(Db, rid, NULL, markersize(RegNode), sizeof(RegNode))) {
		record_t subId;

		AddTail(&RegList, &rn->rn_Node);
		rn->rn_RId = rid;
		rn->rn_RetPath = NULL;
		rn->rn_Alias = NULL;
		rn->rn_Flags &= ~(GF_LOADED | GF_MARKED);
		NewList(&rn->rn_CtlList);
		NewList(&rn->rn_HistList);
		NewList(&rn->rn_AccList);

		/*
		 *  Scan subtree for DBTYPE_REG_RETPATH, _ALIAS,
		 *  for DBTYPE_COMM, _PARAM, and _STAT
		 */

		for (subId = ScanSubRecord(Db, rid, &size, &type); subId; subId = ScanNextRecord(Db, subId, &size, &type)) {
		    switch(type) {
		    case DBTYPE_REG_NAME:
			rn->rn_Node.ln_Name = LoadRecord(Db, subId, NULL, size + 1);
			rn->rn_NId = subId;
			break;
		    case DBTYPE_REG_RETPATH:
			rn->rn_RetPath = LoadRecord(Db, subId, NULL, size + 1);
			rn->rn_RPId = subId;
			break;
		    case DBTYPE_REG_ALIAS:
			rn->rn_Alias = LoadRecord(Db, subId, NULL, size + 1);
			rn->rn_ALId = subId;
			break;
		    case DBTYPE_COMM:
			LoadBMSCommTree(Db, subId, &rn->rn_Comm);
			break;
		    case DBTYPE_PARAM:
			LoadExtRecord(Db, subId, &rn->rn_Param, markersize(Param), sizeof(rn->rn_Param));
			rn->rn_Param.pm_RId = subId;
			break;
		    case DBTYPE_STAT:
			LoadExtRecord(Db, subId, &rn->rn_Stat, markersize(Stat), sizeof(rn->rn_Stat));
			rn->rn_Stat.st_RId = subId;
			break;
		    }
		}
	    }
	}
    }
}

/*
 *  LoadBMSRegTree(rn)
 *
 *  Load the entire registration node tree for a given RegNode.
 */

void
LoadBMSRegTree(rn)
RegNode *rn;
{
    record_t rid;
    long size;
    long type;

    rn->rn_Flags |= GF_LOADED;

    if (rn->rn_RId == 0)
	return;
    for (rid = ScanSubRecord(Db, rn->rn_RId, &size, &type); rid; rid = ScanNextRecord(Db, rid, &size, &type)) {
	CtlNode *cn;

	switch(type) {
	case DBTYPE_CTL:
	    if (cn = LoadRecord(Db, rid, NULL, sizeof(CtlNode))) {
		cn->cn_RId = rid;
		cn->cn_Flags &= ~GF_LOADED;
		NewList(&cn->cn_FileList);
		LoadBMSCtlTree(cn);
		AddTail(&rn->rn_CtlList, &cn->cn_Node);
	    }
	    break;
	case DBTYPE_TXHIST:
	    LoadHistogram(rn, rid);
	    break;
	case DBTYPE_ACCESS:
	    LoadBMSFileAccessTree(rn, rid);
	    break;
	}
    }
}

/*
 *  LoadBMSCtlTree(cn)
 *
 *  Load the entire control node tree for a given CtlNode
 */

void
LoadBMSCtlTree(cn)
CtlNode *cn;
{
    record_t rid;
    long size;
    long type;

    cn->cn_Flags |= GF_LOADED;

    if (cn->cn_RId == 0)
	return;
    for (rid = ScanSubRecord(Db, cn->cn_RId, &size, &type); rid; rid = ScanNextRecord(Db, rid, &size, &type)) {
	FileNode *fn;

	switch(type) {
	case DBTYPE_CTL_NAME:
	    cn->cn_Node.ln_Name = LoadRecord(Db, rid, NULL, size + 1);
	    cn->cn_NId = rid;
	    break;
	case DBTYPE_FILE:
	    if (fn = LoadExtRecord(Db, rid, NULL, markersize(FileNode), sizeof(FileNode))) {
		fn->fn_RId = rid;
		fn->fn_Flags &= ~GF_LOADED;
		NewList(&fn->fn_ChunkList);
		LoadBMSFileTree(fn);
		AddTail(&cn->cn_FileList, &fn->fn_Node);
	    }
	    break;
	case DBTYPE_COMM:
	    LoadBMSCommTree(Db, rid, &cn->cn_Comm);
	    break;
	case DBTYPE_STAT:
	    LoadExtRecord(Db, rid, &cn->cn_Stat, markersize(Stat), sizeof(cn->cn_Stat));
	    cn->cn_Stat.st_RId = rid;
	    break;
	}
    }
}

void
LoadBMSFileAccessTree(rn, rid)
RegNode *rn;
record_t rid;
{
    FileAccess *fa;
    long size;
    long type;
    record_t sid;

    rn->rn_FAId = rid;
    for (rid = ScanSubRecord(Db, rid, &size, &type); rid; rid = ScanNextRecord(Db, rid, &size, &type)) {
	if (type == DBTYPE_ACCESS_NODE) {
	    fa = LoadExtRecord(Db, rid, NULL, markersize(FileAccess), sizeof(FileAccess));
	    fa->fa_RId = rid;
	    AddTail(&rn->rn_AccList, &fa->fa_Node);
	    for (sid = ScanSubRecord(Db, rid, &size, &type); sid; sid = ScanNextRecord(Db, sid, &size, &type)) {
		if (type == DBTYPE_ACCESS_NAME) {
		    fa->fa_Node.ln_Name = LoadRecord(Db, sid, NULL, size + 1);
		    fa->fa_NId = sid;
		}
	    }
	}
    }
}

void
LoadBMSCommTree(db, rid, comm)
DBHan *db;
record_t rid;
Comm *comm;
{
    long size;
    long type;

    LoadExtRecord(db, rid, comm, markersize(Comm), sizeof(Comm));
    comm->com_RId = rid;

    for (rid = ScanSubRecord(db, rid, &size, &type); rid; rid = ScanNextRecord(Db, rid, &size, &type)) {
	switch(type) {
	case DBTYPE_COMM_MSG:
	    comm->com_Msg = LoadRecord(db, rid, NULL, size + 1);
	    comm->com_MId = rid;
	    break;
	}
    }
}

/*
 *  LoadBMSFileTree(fn)
 *
 *  Load the entire file node tree for a given FileNode
 */

void
LoadBMSFileTree(fn)
FileNode *fn;
{
    FileChunk *fc;
    record_t rid;
    long size;
    long type;

    fn->fn_Flags |= GF_LOADED;

    if (fn->fn_RId == 0)
	return;
    for (rid = ScanSubRecord(Db, fn->fn_RId, &size, &type); rid; rid = ScanNextRecord(Db, rid, &size, &type)) {
	switch(type) {
	case DBTYPE_FILE_REMNAME:
	    fn->fn_RemFileName = LoadRecord(Db, rid, NULL, size + 1);
	    fn->fn_RFId = rid;
	    break;
	case DBTYPE_FILE_LOCNAME:
	    fn->fn_LocFileName = LoadRecord(Db, rid, NULL, size + 1);
	    fn->fn_LFId = rid;
	    break;
	case DBTYPE_CHUNK:
	    fn->fn_FCId = rid;
	    LoadBMSChunkTree(fn);
	    break;
	}
    }
}

void
LoadBMSChunkTree(fn)
FileNode *fn;
{
    record_t rid;
    long size;
    long type;
    FileChunk *fc;

    if (fn->fn_FCId) {
	for (rid = ScanSubRecord(Db, fn->fn_FCId, &size, &type); rid; rid = ScanNextRecord(Db, rid, &size, &type)) {
	    if (type == DBTYPE_CHUNK_NODE) {
		fc = LoadExtRecord(Db, rid, NULL, markersize(FileChunk), sizeof(FileChunk));
		fc->fc_RId = rid;
		AddFileChunk(fn, fc);
	    }
	}
    }
}

/*************************************************************************
 *									 *
 *  SaveBMSDataBase()                                                    *
 *									 *
 *  Save modified portions of the BMS database				 *
 *  Delete items marked for deletion					 *
 *									 *
 *************************************************************************/

void
SaveBMSDataBase(void)
{
    RegNode *rn;

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	/*
	 *  Update RegNode
	 */

	if (rn->rn_Flags & GF_DELETED) {
	    if (rn->rn_RId) {
		DeleteRecord(Db, rn->rn_RId);
		rn->rn_RId = 0;
	    }
	    continue;
	}

	if (rn->rn_Flags & GF_MODIFIED) {
	    rn->rn_Flags &= ~GF_MODIFIED;
	    rn->rn_RId = WriteRecord(Db, 0, rn->rn_RId, DBTYPE_REG, rn, markersize(RegNode));

	    if (rn->rn_Node.ln_Name) {
		rn->rn_NId = WriteRecord(Db, rn->rn_RId, rn->rn_NId, DBTYPE_REG_NAME, rn->rn_Node.ln_Name, strlen(rn->rn_Node.ln_Name));
	    } else if (rn->rn_NId) {
		DeleteRecord(Db, rn->rn_NId);
		rn->rn_NId = 0;
	    }

	    if (rn->rn_RetPath) {
		rn->rn_RPId = WriteRecord(Db, rn->rn_RId, rn->rn_RPId, DBTYPE_REG_RETPATH, rn->rn_RetPath, strlen(rn->rn_RetPath));
	    } else if (rn->rn_RPId) {
		DeleteRecord(Db, rn->rn_RPId);
		rn->rn_RPId = 0;
	    }

	    if (rn->rn_Alias) {
		rn->rn_ALId = WriteRecord(Db, rn->rn_RId, rn->rn_ALId, DBTYPE_REG_ALIAS, rn->rn_Alias, strlen(rn->rn_Alias));
	    } else if (rn->rn_ALId) {
		DeleteRecord(Db, rn->rn_ALId);
		rn->rn_ALId = 0;
	    }
	}

	/*
	 *  Update Comm
	 */

	if (rn->rn_Comm.com_Flags & GF_DELETED) {
	    if (rn->rn_Comm.com_RId) {
		DeleteRecord(Db, rn->rn_Comm.com_RId);
		rn->rn_Comm.com_RId = 0;
		rn->rn_Comm.com_MId = 0;
	    }
	} else if (rn->rn_Comm.com_Flags & GF_MODIFIED) {
	    rn->rn_Comm.com_Flags &= ~GF_MODIFIED;
	    rn->rn_Comm.com_RId = WriteRecord(Db, rn->rn_RId, rn->rn_Comm.com_RId, DBTYPE_COMM, &rn->rn_Comm, markersize(Comm));
	    if (rn->rn_Comm.com_Msg)
		rn->rn_Comm.com_MId = WriteRecord(Db, rn->rn_Comm.com_RId, rn->rn_Comm.com_MId, DBTYPE_COMM_MSG, rn->rn_Comm.com_Msg, strlen(rn->rn_Comm.com_Msg));
	}

	/*
	 *  Update Param
	 */

	if (rn->rn_Param.pm_Flags & GF_DELETED) {
	    if (rn->rn_Param.pm_RId) {
		DeleteRecord(Db, rn->rn_Param.pm_RId);
		rn->rn_Param.pm_RId = 0;
	    }
	} else if (rn->rn_Param.pm_Flags & GF_MODIFIED) {
	    rn->rn_Param.pm_Flags &= ~GF_MODIFIED;
	    rn->rn_Param.pm_RId = WriteRecord(Db, rn->rn_RId, rn->rn_Param.pm_RId, DBTYPE_PARAM, &rn->rn_Param, markersize(Param));
	}

	/*
	 *  Update Stat
	 */

	if (rn->rn_Stat.st_Flags & GF_DELETED) {
	    if (rn->rn_Stat.st_RId) {
		DeleteRecord(Db, rn->rn_Stat.st_RId);
		rn->rn_Stat.st_RId = 0;
	    }
	} else if (rn->rn_Stat.st_Flags & GF_MODIFIED) {
	    rn->rn_Stat.st_Flags &= ~GF_MODIFIED;
	    rn->rn_Stat.st_RId = WriteRecord(Db, rn->rn_RId, rn->rn_Stat.st_RId, DBTYPE_STAT, &rn->rn_Stat, markersize(Stat));
	}

	/*
	 *  Scan CtlList for CtlNodes
	 */

	SaveBMSCtlList(&rn->rn_CtlList, rn->rn_RId);

	/*
	 *  Scan HistList for HistNodes
	 */

	SaveHistogram(rn);
	SaveBMSFileAccessTree(rn);
    }
}

void
SaveBMSCtlList(list, reg_id)
List *list;
record_t reg_id;
{
    CtlNode *cn;

    for (cn = GetHead(list); cn; cn = GetSucc(&cn->cn_Node)) {
	if (cn->cn_Flags & GF_DELETED) {
	    if (cn->cn_RId) {
		DeleteRecord(Db, cn->cn_RId);
		cn->cn_RId = 0;
	    }
	    continue;
	}

	if (cn->cn_Flags & GF_MODIFIED) {
	    cn->cn_Flags &= ~GF_MODIFIED;
	    cn->cn_RId = WriteRecord(Db, reg_id, cn->cn_RId, DBTYPE_CTL, cn, markersize(CtlNode));

	    if (cn->cn_Node.ln_Name) {
		cn->cn_NId = WriteRecord(Db, cn->cn_RId, cn->cn_NId, DBTYPE_CTL_NAME, cn->cn_Node.ln_Name, strlen(cn->cn_Node.ln_Name));
	    } else if (cn->cn_NId) {
		DeleteRecord(Db, cn->cn_NId);
		cn->cn_NId = 0;
	    }
	}

	/*
	 *  Update COMM
	 */

	if (cn->cn_Comm.com_Flags & GF_DELETED) {
	    if (cn->cn_Comm.com_RId) {
		DeleteRecord(Db, cn->cn_Comm.com_RId);
		cn->cn_Comm.com_RId = 0;
		cn->cn_Comm.com_MId = 0;
	    }
	} else if (cn->cn_Comm.com_Flags & GF_MODIFIED) {
	    cn->cn_Comm.com_Flags &= ~GF_MODIFIED;
	    cn->cn_Comm.com_RId = WriteRecord(Db, cn->cn_RId, cn->cn_Comm.com_RId, DBTYPE_COMM, &cn->cn_Comm, markersize(Comm));
	    if (cn->cn_Comm.com_Msg)
		cn->cn_Comm.com_MId = WriteRecord(Db, cn->cn_Comm.com_RId, cn->cn_Comm.com_MId, DBTYPE_COMM_MSG, cn->cn_Comm.com_Msg, strlen(cn->cn_Comm.com_Msg));
	}

	/*
	 *  Update Stat
	 */

	if (cn->cn_Stat.st_Flags & GF_DELETED) {
	    if (cn->cn_Stat.st_RId) {
		DeleteRecord(Db, cn->cn_Stat.st_RId);
		cn->cn_Stat.st_RId = 0;
	    }
	} else if (cn->cn_Stat.st_Flags & GF_MODIFIED) {
	    cn->cn_Stat.st_Flags &= ~GF_MODIFIED;
	    cn->cn_Stat.st_RId = WriteRecord(Db, cn->cn_RId, cn->cn_Stat.st_RId, DBTYPE_STAT, &cn->cn_Stat, markersize(Stat));
	}

	SaveBMSFileList(&cn->cn_FileList, cn->cn_RId);
    }
}

void
SaveBMSFileAccessTree(rn)
RegNode *rn;
{
    FileAccess *fa;

    for (fa = GetHead(&rn->rn_AccList); fa; fa = GetSucc(&fa->fa_Node)) {
	if (rn->rn_FAId == 0)
	    rn->rn_FAId = WriteRecord(Db, rn->rn_RId, 0, DBTYPE_ACCESS, NULL, 0);
	if (fa->fa_Flags & GF_DELETED) {
	    if (fa->fa_RId) {
		DeleteRecord(Db, fa->fa_RId);
		fa->fa_RId = 0;
	    }
	} else if (fa->fa_Flags & GF_MODIFIED) {
	    fa->fa_Flags &= ~GF_MODIFIED;
	    fa->fa_RId = WriteRecord(Db, rn->rn_FAId, fa->fa_RId, DBTYPE_ACCESS_NODE, fa, markersize(FileAccess));

	    if (fa->fa_Node.ln_Name) {
		fa->fa_NId = WriteRecord(Db, fa->fa_RId, fa->fa_NId, DBTYPE_ACCESS_NAME, fa->fa_Node.ln_Name, strlen(fa->fa_Node.ln_Name));
	    } else if (fa->fa_NId) {
		DeleteRecord(Db, fa->fa_NId);
		fa->fa_NId = 0;
	    }
	}
    }
}

void
SaveBMSFileList(list, ctl_id)
List *list;
record_t ctl_id;
{
    FileNode *fn;

    for (fn = GetHead(list); fn; fn = GetSucc(&fn->fn_Node)) {
	if (fn->fn_Flags & GF_DELETED) {
	    if (fn->fn_RId) {
		DeleteRecord(Db, fn->fn_RId);
		fn->fn_RId = 0;
	    }
	    continue;
	}
	if (fn->fn_Flags & GF_MODIFIED) {
	    fn->fn_Flags &= ~GF_MODIFIED;
	    fn->fn_RId = WriteRecord(Db, ctl_id, fn->fn_RId, DBTYPE_FILE, fn, markersize(FileNode));

	    if (fn->fn_RemFileName) {
		fn->fn_RFId = WriteRecord(Db, fn->fn_RId, fn->fn_RFId, DBTYPE_FILE_REMNAME, fn->fn_RemFileName, strlen(fn->fn_RemFileName));
	    } else if (fn->fn_RFId) {
		DeleteRecord(Db, fn->fn_RFId);
		fn->fn_RFId = 0;
	    }

	    if (fn->fn_LocFileName) {
		fn->fn_LFId = WriteRecord(Db, fn->fn_RId, fn->fn_LFId, DBTYPE_FILE_LOCNAME, fn->fn_LocFileName, strlen(fn->fn_LocFileName));
	    } else if (fn->fn_LFId) {
		DeleteRecord(Db, fn->fn_LFId);
		fn->fn_LFId = 0;
	    }

	    if (fn->fn_Comment) {
		fn->fn_CMId = WriteRecord(Db, fn->fn_RId, fn->fn_CMId, DBTYPE_FILE_COMMENT, fn->fn_Comment, strlen(fn->fn_Comment));
	    } else if (fn->fn_CMId) {
		DeleteRecord(Db, fn->fn_CMId);
		fn->fn_CMId = 0;
	    }
	}
	SaveBMSChunkList(fn);
    }
}

void
SaveBMSChunkList(fn)
FileNode *fn;
{
    FileChunk *fc;

    for (fc = GetHead(&fn->fn_ChunkList); fc; fc = GetSucc(&fc->fc_Node)) {

	if (fn->fn_FCId == 0)
	    fn->fn_FCId = WriteRecord(Db, fn->fn_RId, fn->fn_FCId, DBTYPE_CHUNK, NULL, 0);

	if (fc->fc_Flags & GF_DELETED) {
	    if (fc->fc_RId) {
		DeleteRecord(Db, fc->fc_RId);
		fc->fc_RId = 0;
	    }
	} else if (fc->fc_Flags & GF_MODIFIED) {
	    fc->fc_Flags &= ~GF_MODIFIED;
	    fc->fc_RId = WriteRecord(Db, fn->fn_FCId, fc->fc_RId, DBTYPE_CHUNK_NODE, fc, markersize(FileChunk));
	}
    }
}


/*************************************************************************
 *									 *
 *  UnLoadBMSDataBase()                                                  *
 *									 *
 *  Unload the database, free all structures				 *
 *  (note: You must call SaveBMSDataBase() first, we do not)             *
 *									 *
 *************************************************************************/

void
UnLoadBMSDataBase(void)
{
    RegNode *rn;

    while (rn = GetHead(&RegList)) {
	UnLoadBMSRegTree(rn);

	Remove(&rn->rn_Node);

	if (rn->rn_Node.ln_Name)
	    zfree(rn->rn_Node.ln_Name);
	if (rn->rn_RetPath)
	    zfree(rn->rn_RetPath);
	if (rn->rn_Alias)
	    zfree(rn->rn_Alias);

	zfree(rn);
    }
}

void
UnLoadBMSRegTree(rn)
RegNode *rn;
{
    {
	CtlNode *cn;

	while (cn = GetHead(&rn->rn_CtlList))
	    UnLoadBMSCtlTree(cn);
    }
    {
	FileAccess *fa;

	while (fa = GetHead(&rn->rn_AccList))
	    UnLoadBMSFileAccessTree(fa);
    }

    UnLoadHistogram(rn);
}

void
UnLoadBMSCtlTree(cn)
CtlNode *cn;
{
    FileNode *fn;

    Remove(&cn->cn_Node);

    if (cn->cn_Node.ln_Name)
	zfree(cn->cn_Node.ln_Name);

    while (fn = GetHead(&cn->cn_FileList))
	UnLoadBMSFileTree(fn);
    zfree(cn);
}

void
UnLoadBMSFileAccessTree(fa)
FileAccess *fa;
{
    Remove(&fa->fa_Node);

    if (fa->fa_Node.ln_Name)
	zfree(fa->fa_Node.ln_Name);
    zfree(fa);
}

void
UnLoadBMSFileTree(fn)
FileNode *fn;
{
    FileChunk *fc;

    Remove(&fn->fn_Node);

    while (fc = RemHead(&fn->fn_ChunkList))
	zfree(fc);

    if (fn->fn_Node.ln_Name)
	zfree(fn->fn_Node.ln_Name);
    if (fn->fn_LocFileName)
	zfree(fn->fn_LocFileName);
    if (fn->fn_RemFileName)
	zfree(fn->fn_RemFileName);

    zfree(fn);
}

/*************************************************************************
 *									 *
 *  AllocRegNode()                                                       *
 *									 *
 *  Allocate a registration node					 *
 *									 *
 *************************************************************************/

RegNode *
AllocRegNode()
{
    RegNode *rn = zalloc(sizeof(RegNode));

    rn->rn_Flags |= GF_MODIFIED;
    rn->rn_Comm.com_Flags |= GF_MODIFIED;
    rn->rn_Param.pm_Flags |= GF_MODIFIED;
    rn->rn_Stat.st_Flags  |= GF_MODIFIED;
    NewList(&rn->rn_CtlList);
    NewList(&rn->rn_HistList);
    NewList(&rn->rn_AccList);
    AddTail(&RegList, &rn->rn_Node);

    return(rn);
}

CtlNode *
AllocCtlNode(rn)
RegNode *rn;
{
    CtlNode *cn = zalloc(sizeof(CtlNode));

    cn->cn_Flags |= GF_MODIFIED;
    NewList(&cn->cn_FileList);
    AddTail(&rn->rn_CtlList, &cn->cn_Node);

    return(cn);
}

FileNode *
AllocFileNode(cn)
CtlNode *cn;
{
    FileNode *fn = zalloc(sizeof(FileNode));

    fn->fn_Flags |= GF_MODIFIED;
    cn->cn_Flags |= GF_MODIFIED;
    NewList(&fn->fn_ChunkList);

    AddTail(&cn->cn_FileList, &fn->fn_Node);

    return(fn);
}

FileChunk *
AllocFileChunk(fn, offset, bytes)
FileNode *fn;
long offset;
long bytes;
{
    FileChunk *fc = zalloc(sizeof(FileChunk));

    fc->fc_Flags |= GF_MODIFIED;
    fc->fc_Bytes = bytes;
    fc->fc_Offset = offset;
    AddFileChunk(fn, fc);
    return(fc);
}

FileAccess *
AllocFileAccess(rn, name)
RegNode *rn;
const char *name;
{
    FileAccess *fa = zalloc(sizeof(FileAccess));

    fa->fa_Flags |= GF_MODIFIED;
    fa->fa_Node.ln_Name = zstrdup(name);
    AddTail(&rn->rn_AccList, &fa->fa_Node);
    return(fa);
}

void
AddFileChunk(fn, fc)
FileNode *fn;
FileChunk *fc;
{
    FileChunk *fscan;

    for (fscan = GetHead(&fn->fn_ChunkList); fscan; fscan = GetSucc(&fscan->fc_Node)) {
	if (fc->fc_Flags & GF_DELETED)
	    continue;
	if (fc->fc_Offset <= fscan->fc_Offset)
	    break;
    }
    if (fscan == NULL)
	AddTail(&fn->fn_ChunkList, &fc->fc_Node);
    else
	Insert(&fn->fn_ChunkList, &fc->fc_Node, GetPred(&fscan->fc_Node));
}



