
/*
 *  DATABASE.C
 *
 *  This module provides hierarchical database handling routines as well
 *  as automatic byte ordering for the database superstructure itself.
 *  Byte ordering for the contents of objects is not automatic since the
 *  database has no way of knowing the structural format of said
 *  objects.  However, the BMS system always stuffs ASCII data for
 *  BMS databases transfered over email.  Therefore, it should be
 *  compatible.
 *
 *  db = OpenDataBase(fileName, modes);
 *  recordid = ScanRecord(db, rid, &size, &type);
 *  recordid = ScanNextRecord(db, rid,  &size, &type);
 *  recordid = ScanSubRecord(db, rid, &size, &type);
 *  (void) CloseDataBase(db)
 *
 *  newrid = WriteRecord(db, parent_id, rid/0, type, buf, bytes)
 *  (void)   MoveRecord(db, parent_id, rid, newparent_id)
 *  newrid = ModifyRecord(db, recordid, type, buf, bytes)
 *  DeleteRecord(db, recordid)
 *  rid = AppendRecord(db, recordid_tree, type, buf, bytes)
 *
 *  ptr = LoadRecord(db, recordid, buf/NULL, bytes)
 *  ptr = LoadExtRecord(db, recordid, buf/NULL, dbbytes, struct_size)
 *  (void) CleanDataBase(fileName)
 *
 *  note: if more bytes are specified then exist in the record, the extra
 *  space will be zerod.
 */

#include "defs.h"

#ifdef TEST
#define zalloc(bytes)   calloc(bytes,1)
#define zfree(ptr)      free(ptr)
#define DebugOpt    1
#endif

Prototype DBHan *OpenDataBase(const char *, const char *);
Prototype record_t ScanRecord(DBHan *, long,  long *, long *);
Prototype record_t ScanNextRecord(DBHan *, long,  long *, long *);
Prototype record_t ScanSubRecord(DBHan *, long, long *, long *);
Prototype void CloseDataBase(DBHan *);

Prototype record_t WriteRecord(DBHan *, long, long, long, void *, long);
Prototype record_t MoveRecord(DBHan *, record_t, record_t, record_t);
Prototype record_t ModifyRecord(DBHan *, long, long, void *, long);
Prototype void DeleteRecord(DBHan *, long);
Prototype record_t AppendRecord(DBHan *, long, long, void *, long);
Prototype void *LoadRecord(DBHan *, long, void *, long);
Prototype void *LoadExtRecord(DBHan *, long, void *, long, long);

Prototype void CleanDataBase(const char *);

int CopyDataBase(DBHan *, DBHan *, record_t, record_t);
void DeleteTree(DBHan *, record_t);
int ObtainRecord(DBHan *, record_t);
void FixByteOrdering(DBHan *, record_t, long);
unsigned long Unscramble(unsigned long, unsigned long);

/*
 *  OPENDATABASE()
 *
 *  Open (modes == 1006) or Create (modes == 1005) a database.  If the
 *  byte ordering is incorrect it will be fixed in place, modifying
 *  the database file.
 */

DBHan *
OpenDataBase(fileName, mode)
const char *fileName;
const char *mode;
{
    DBHan *db = NULL;
    FILE *fi;
    short create = 0;
    int flags = 0;

    if (mode[0] == 'w') {
	flags = 1006;
	create = 1;
    } else {
	flags = 1005;
    }

    if (fi = copen(fileName, flags)) {
	db = zalloc(sizeof(DBHan));
	db->db_Fi = fi;
	db->db_AppRoot = -1;		/*  clear append cache	*/
	db->db_RId = -1;		/*  clear rid cache	*/
	db->db_Head.db_Version= 1;
	db->db_Head.db_Endian = 0x04030201;

	if (create)
	    cwrite(db->db_Fi, &db->db_Head, sizeof(db->db_Head));
	else
	    cread(db->db_Fi, &db->db_Head, sizeof(db->db_Head));

	/*
	 *  Check byte ordering for database headers.  Upward compatible
	 *  (field was 0 in older versions of BMS).  If byte ordering
	 *  is wrong then we fix it in place.
	 *
	 *  This feature is mainly for temporary databases which are then
	 *  transmitted across a network and is unrelated to any permanent
	 *  machine databases that are not intended to be transmitted
	 *  across a network.
	 */

	if (db->db_Head.db_Endian && db->db_Head.db_Endian != 0x04030201) {
	    FixByteOrdering(db, 0, db->db_Head.db_Endian);
	}
    }
    return(db);
}

/*
 *  SCANRECORD()
 *
 *  Scan a specific record.  This is the only call for which rid == 0 is
 *  special cased.  Rather then return the root node we return the root
 *  node's subtree which is the real beginning of the database.
 *
 *  The special case is the subtree of the root node which must be
 *  0 for the empty-database case (so append works), but may not be in
 *  other cases.
 *
 *  0 is returned only for the case of an empty database.
 */

record_t
ScanRecord(db, rid, psize, ptype)
DBHan *db;
record_t rid;
long *psize;
long *ptype;
{
    if (rid == 0) {
	ObtainRecord(db, 0);
	if ((rid = db->db_Record.dr_SubTree) == 0)
	    rid = sizeof(db->db_Head);
    }
    if (db->db_RId == rid) {
	*ptype = db->db_Record.dr_Type;
	*psize = db->db_Record.dr_Size;
    } else {
	cseek(db->db_Fi, rid);
	if (cread(db->db_Fi, &db->db_Record, sizeof(DBRecord)) == sizeof(DBRecord)) {
	    *ptype = db->db_Record.dr_Type;
	    *psize = db->db_Record.dr_Size;
	    db->db_RId = rid;
	} else {
	    db->db_RId = 0;
	    rid = 0;
	}
    }
    if (DebugOpt)
	printf("SR %d type=%08lx size=%08lx\n", rid, *ptype, *psize);
    return(rid);
}

/*
 *  SCANNEXTRECORD()
 *
 *  Scan the next linear record at this level of the heirarchy, return
 *  a new record number or 0 if there are no more records.
 */

record_t
ScanNextRecord(db, rid, psize, ptype)
DBHan *db;
record_t rid;
long *psize;
long *ptype;
{
    if (ObtainRecord(db, rid) == 0 || db->db_Record.dr_Chain == 0)
	return(0);

    /*
     *	Use information in current record to scan next record
     */

    return(ScanRecord(db, db->db_Record.dr_Chain, psize, ptype));
}

/*
 *  SCANSUBRECORD()
 *
 *  Scan the first record of the subtree under this record, used to
 *  decend a database heirarchy.  0 is returned if there is no
 *  subheirarchy under this record.
 */

record_t
ScanSubRecord(db, rid, psize, ptype)
DBHan *db;
record_t rid;
long *psize;
long *ptype;
{
    if (ObtainRecord(db, rid) == 0 || db->db_Record.dr_SubTree == 0)
	return(0);

    /*
     *	Use information in current record to scan next record
     */

    return(ScanRecord(db, db->db_Record.dr_SubTree, psize, ptype));
}

/*
 *  CLOSEDATABASE()
 *
 *  Close the database.  XXX Eventually this call will cleanup the
 *  database every so often.
 */

void
CloseDataBase(db)
DBHan *db;
{
    cclose(db->db_Fi);
    zfree(db);
}

/*
 *  WRITERECORD()
 *
 *  This routine modifies or appends a record.	If rid is 0 then a new
 *  record is created in the parent's (pid) subtree, appended to the
 *  end.
 *
 *  If rid is not zero then that specific record is overwritten.
 */

record_t
WriteRecord(db, pid, rid, type, buf, bytes)
DBHan *db;
record_t pid;
record_t rid;
long type;
void *buf;
long bytes;
{
    if (DebugOpt)
	printf("WriteRecord pid=%-5d rid=%-5d type=%08lx\n", pid, rid, type);
    if (rid == 0)
	return(AppendRecord(db, pid, type, buf, bytes));
    else
	return(ModifyRecord(db, rid, type, buf, bytes));
}


/*
 *  MODIFYRECORD()
 *
 *  This routine returns the rid of the new record which may or may not
 *  be different then the rid of the old record.  If different, the
 *  old record is deleted.  THE OLD RECORD IDENTIFIER (RID) IS INVALID
 *
 *  This routine may create deadspace in the database.
 */

record_t
ModifyRecord(db, rid, type, buf, bytes)
DBHan *db;
record_t rid;
long type;
void *buf;
long bytes;
{
    /*
     *	Obtain the record.  Cannot modify the root record!
     */

    if (ObtainRecord(db, rid) == 0)
	return(0);

    /*
     *	Update the record
     */

    if (bytes <= db->db_Record.dr_Size) {
	/*
	 *  Data fits
	 */

	cseek(db->db_Fi, db->db_RId);
	db->db_Record.dr_Type = type;
	db->db_Record.dr_Size = bytes;
	cwrite(db->db_Fi, &db->db_Record, sizeof(DBRecord));
	cwrite(db->db_Fi, buf, bytes);
    } else {
	/*
	 *  Data does not fit
	 */
	DBRecord new;
	record_t newChain;

	cseek(db->db_Fi, csize(db->db_Fi));
	newChain = csize(db->db_Fi);
	new.dr_Type = type;
	new.dr_Size = bytes;
	new.dr_Chain = db->db_Record.dr_Chain;
	new.dr_SubTree= db->db_Record.dr_SubTree;
	cwrite(db->db_Fi, &new, sizeof(DBRecord));
	cwrite(db->db_Fi, buf, bytes);

	cseek(db->db_Fi, db->db_RId);
	db->db_Record.dr_Type |= 0x80000000;	/*  delete record   */
	db->db_Record.dr_Chain = newChain;
	db->db_Record.dr_SubTree = 0;
	cwrite(db->db_Fi, &db->db_Record, sizeof(DBRecord));
	db->db_Record = new;
	db->db_RId = newChain;
    }
    return(db->db_RId);
}

/*
 *  MOVERECORD()
 *
 *  This routine moves a record and its subtree, if any, to a different
 *  location in the database.  The move is done in place and no dead
 *  space is created by the call.
 *
 *  par_rid	-   record id of parent of the record to be moved
 *  rid 	-   record to be moved
 *  newpar_rid	-   new parent for the record.	rid is moved into
 *		    newpar_rid's subtree.
 *
 *  WARNING!  Moving a record into a subtree of itself is ILLEGAL and
 *  not supported.
 */

record_t
MoveRecord(db, par_rid, rid, newpar_rid)
DBHan *db;
record_t par_rid;
record_t rid;
record_t newpar_rid;
{
    record_t xid;
    record_t nid;

    /*
     *	Get the parent record
     */

    if (ObtainRecord(db, rid) == 0)
	return(0);
    nid = db->db_Record.dr_Chain;   /*	chain link  */

    if (ObtainRecord(db, par_rid) == 0)
	return(0);

    /*
     *	Start with the sub-tree looking for a chain link to rid,
     *	unlink rid from the parent directory
     */

    if ((xid = db->db_Record.dr_SubTree) == rid) {
	db->db_Record.dr_SubTree = nid;
    } else {
	for (;;) {
	    ObtainRecord(db, xid);
	    if (db->db_Record.dr_Chain == 0 || db->db_Record.dr_Chain == rid)
		break;
	}
	if (db->db_Record.dr_Chain == 0)    /*  BAD PARENT WAS SPECIFIED! */
	    return(0);
	db->db_Record.dr_Chain = nid;
    }
    cseek(db->db_Fi, db->db_RId);
    cwrite(db->db_Fi, &db->db_Record, sizeof(DBRecord));

    /*
     *	link rid under newpar_rid, link in front XXX ???
     */

    ObtainRecord(db, newpar_rid);
    nid = db->db_Record.dr_SubTree;
    db->db_Record.dr_SubTree = rid;

    cseek(db->db_Fi, db->db_RId);
    cwrite(db->db_Fi, &db->db_Record, sizeof(DBRecord));

    ObtainRecord(db, rid);
    db->db_Record.dr_Chain = nid;

    cseek(db->db_Fi, db->db_RId);
    cwrite(db->db_Fi, &db->db_Record, sizeof(DBRecord));

    return(xid);
}

/*
 *  DELETERECORD()
 *
 *  This routine deletes a record.  The record becomes dead space.
 *  Deletion consists of marking the record for deletion by setting
 *  bit 31 of the TYPE field.  These records will still show up in
 *  a scan but should be ignored.
 *
 *  Any subtree will also be deleted.  The physical space taken up
 *  by dead records and fragments is reclaimed by CleanDataBase()
 */

void
DeleteRecord(db, rid)
DBHan *db;
record_t rid;
{
    record_t subTree;

    /*
     *	Obtain the record, cannot delete the root record!
     */

    if (ObtainRecord(db, rid) == 0)
	return;

    /*
     *	Mark record as deleted
     */

    cseek(db->db_Fi, db->db_RId);
    subTree = db->db_Record.dr_SubTree;
    db->db_Record.dr_Type |= 0x80000000;
    db->db_Record.dr_SubTree = 0;
    cwrite(db->db_Fi, &db->db_Record, sizeof(DBRecord));

    if (subTree)
	DeleteTree(db, subTree);
}

/*
 *  APPENDRECORD
 *
 *  Append a new record to the subtree of rid.	rid may be 0 which
 *  indicates appending a new record to the top level of the database.
 */

record_t
AppendRecord(db, root, type, buf, bytes)
DBHan *db;
record_t root;
long type;
void *buf;
long bytes;
{
    record_t rid;
    record_t newRid;
    DBRecord new;

    /*
     *	Find last record in subtree of root so we can append our new record.
     */

    if (root == db->db_AppRoot) {
	rid = db->db_AppLast;
    } else if (ObtainRecord(db, root) || root == 0) {
	rid = db->db_Record.dr_SubTree;
    } else {
	return(0);
    }

    /*
     *	Scan records until we find the end of the chain, so we can modify
     *	the dr_Chain entry of the last record.	If the chain is empty
     *	we must modify the dr_SubTree entry of the root.
     */

    if (rid) {
	record_t next;
	long size;
	long type;

	while (next = ScanNextRecord(db, rid, &size, &type)) {
	    rid = next;
	}
    }

    /*
     *	Append the new record
     */

    cseek(db->db_Fi, csize(db->db_Fi));
    newRid = csize(db->db_Fi);
    new.dr_Type = type;
    new.dr_Size = bytes;
    new.dr_Chain = 0;
    new.dr_SubTree= 0;
    cwrite(db->db_Fi, &new, sizeof(DBRecord));
    cwrite(db->db_Fi, buf, bytes);

    /*
     *	Modify the chain entry or subtree of the root depending
     */

    if (rid) {
	ObtainRecord(db, rid);
	cseek(db->db_Fi, db->db_RId);
	db->db_Record.dr_Chain = newRid;
	cwrite(db->db_Fi, &db->db_Record, sizeof(DBRecord));
    } else {
	ObtainRecord(db, root);
	cseek(db->db_Fi, db->db_RId);
	db->db_Record.dr_SubTree = newRid;
	cwrite(db->db_Fi, &db->db_Record, sizeof(DBRecord));
    }
    return(newRid);
}

/*
 *  LOADRECORD()
 *
 *  Load the data associated with a record.  If the data buffer is larger
 *  then the data in the record the buffer will be zero padded.  If the
 *  data buffer is smaller then the data in the record only <bytes>
 *  bytes are copied.
 *
 *  zero fill only occurs if no buffer is specified (buf = NULL), when
 *  we do the allocating of the buffer.
 */

void *
LoadRecord(db, rid, buf, bytes)
DBHan *db;
record_t rid;
void *buf;
long bytes;
{
    if (ObtainRecord(db, rid) == 0)
	return(NULL);
    if (buf == NULL)
	buf = zalloc(bytes);
    cseek(db->db_Fi, db->db_RId + sizeof(DBRecord));
    if (bytes > db->db_Record.dr_Size)
	bytes = db->db_Record.dr_Size;
    cread(db->db_Fi, buf, bytes);
    return(buf);
}

/*
 *  LOADEXTRECORD()
 *
 *  This is similar to LoadRecord() but has an additional argument.  You
 *  specify the amount of data to be read from the record and the size
 *  of the memory structure separately.
 *
 *  This provides for compatibility between versions of BMS by not
 *  'spilling over' database data into non-database areas of a memory
 *  structure.
 *
 *  zero fill only occurs if no buffer is specified (buf = NULL), when
 *  we do the allocating of the buffer.
 */

void *
LoadExtRecord(db, rid, buf, bytes, struct_size)
DBHan *db;
record_t rid;
void *buf;
long bytes;
long struct_size;
{
    if (ObtainRecord(db, rid) == 0)
	return(NULL);
    if (buf == NULL)
	buf = zalloc(struct_size);
    cseek(db->db_Fi, db->db_RId + sizeof(DBRecord));
    if (bytes > db->db_Record.dr_Size)
	bytes = db->db_Record.dr_Size;
    cread(db->db_Fi, buf, bytes);
    return(buf);
}


/*
 *  CLEANDATABASE()
 *
 *  Cleanup a database.  The database must not be open at this time.
 *  Cleanup operation consists of scanning the entire database and
 *  rewriting it to a temporary which is then renamed back to the
 *  original.
 */

void
CleanDataBase(fileName)
const char *fileName;
{
    char *tmpFileName = malloc(strlen(fileName) + 32);
    sprintf(tmpFileName, "%s.TMP", fileName);
    DBHan *dbs;
    DBHan *dbd;
    long r = -1;

    if (dbs = OpenDataBase(fileName, "r")) {
	if (dbd = OpenDataBase(tmpFileName, "w")) {
	    r = CopyDataBase(dbs, dbd, 0, 0);
	    CloseDataBase(dbd);
	}
	CloseDataBase(dbs);
    }
    if (r == 0) {
	remove(fileName);
	rename(tmpFileName, fileName);
    } else {
	remove(tmpFileName);
    }
    free(tmpFileName);
}

int
CopyDataBase(dbs, dbd, srid, drid)
DBHan *dbs;
DBHan *dbd;
record_t srid;
record_t drid;
{
    long size;
    long type;
    record_t rid;
    record_t did;
    void *ptr;
    long r = 0;

    for (rid = ScanRecord(dbs, srid, &size, &type); r == 0 && rid; rid = ScanNextRecord(dbs, rid, &size, &type)) {
	record_t subId;

	/*
	 *  Copy the record and subtree.  Do not copy records marked for
	 *  deletion.
	 */

	if (type & 0x80000000)
	    continue;

	did = 0;
	if (size) {
	    if (ptr = LoadRecord(dbs, rid, NULL, size)) {
		did = AppendRecord(dbd, drid, type, ptr, size);
		zfree(ptr);
	    } else {
		r = -1;
		break;
	    }
	} else {
	    did = AppendRecord(dbd, drid, type, NULL, 0);
	}

	if (subId = ScanSubRecord(dbs, rid, &size, &type))
	    r = CopyDataBase(dbs, dbd, subId, did);
    }
    return(r);
}

void
DeleteTree(db, rid)
DBHan *db;
record_t rid;
{
    long size;
    long type;

    while (rid) {
	DeleteRecord(db, rid);
	rid = ScanNextRecord(db, rid, &size, &type);
    }
}

int
ObtainRecord(db, rid)
DBHan *db;
record_t rid;
{
    /*
     *	Obtain this record (rid) to get dr_SubTree entry
     */

    if (rid != db->db_RId) {
	cseek(db->db_Fi, rid);
	if (cread(db->db_Fi, &db->db_Record, sizeof(DBRecord)) == sizeof(DBRecord)) {
	    db->db_RId = rid;
	} else {
	    db->db_RId = 0;
	}
    }
    return(db->db_RId);
}

/*
 *  FIXBYTEORDERING()   (local)
 *
 *  This routine scans the entire database fixing database headers in
 *  place when it is determined to have incorrect byte ordering.
 */

void
FixByteOrdering(db, rid, endi)
DBHan *db;
record_t rid;
long endi;
{
    record_t sid;
    long scram_size;
    long scram_type;

    if (rid == 0) {
	db->db_Head.db_Version = Unscramble(endi, db->db_Head.db_Version);
	db->db_Head.db_Endian  = Unscramble(endi, db->db_Head.db_Endian);
	db->db_Head.db_Reserved= Unscramble(endi, db->db_Head.db_Reserved);
	db->db_Head.db_SubTree = Unscramble(endi, db->db_Head.db_SubTree);
	cseek(db->db_Fi, 0);
	cwrite(db->db_Fi, &db->db_Head, sizeof(db->db_Head));
    }

    /*
     *	Fix records as we scan them
     */

    for (rid = ScanRecord(db, rid, &scram_size, &scram_type); rid; rid = ScanNextRecord(db, rid, &scram_size, &scram_type)) {
	ObtainRecord(db, rid);
	db->db_Record.dr_Type	= Unscramble(endi, db->db_Record.dr_Type);
	db->db_Record.dr_Size	= Unscramble(endi, db->db_Record.dr_Size);
	db->db_Record.dr_Chain	= Unscramble(endi, db->db_Record.dr_Chain);
	db->db_Record.dr_SubTree= Unscramble(endi, db->db_Record.dr_SubTree);
	cseek(db->db_Fi, rid);
	cwrite(db->db_Fi, &db->db_Record, sizeof(db->db_Record));
	if (db->db_Record.dr_SubTree)
	    FixByteOrdering(db, db->db_Record.dr_SubTree, endi);
    }
}

/*
 *  UNSCRAMBLE()
 *
 *  endi should be 0x04030201 .. if it isn't, then we have a byte ordering
 *  problem which must be unscrambled.
 */

unsigned long
Unscramble(endi, value)
unsigned long endi;
unsigned long value;
{
    long r;

    switch(endi) {
    case 0x04030201:	/*  ordering ok 	*/
	r = value;
	break;
    case 0x01020304:	/*  ordering backwords	*/
	r = ((value & 0x000000FF) << 24) |
	    ((value & 0x0000FF00) << 8) |
	    ((value & 0x00FF0000) >> 8) |
	    ((value & 0xFF000000) >> 24);
	break;
    default:		/*  ordering very wierd */
	r = 0;
						/*  4	3   2	1   */
	switch(endi & 0xFF000000) {
	case 0x01000000:			/*  1	x   x	x   */
	    r |= (value >> 24) & 0x000000FF;
	    break;
	case 0x02000000:			/*  2	x   x	x   */
	    r |= (value >> 16) & 0x0000FF00;
	    break;
	case 0x03000000:			/*  3	x   x	x   */
	    r |= (value >>  8) & 0x00FF0000;
	    break;
	case 0x04000000:			/*  4	x   x	x   */
	    r |= value & 0xFF000000;
	    break;
	}
						/*  4	3   2	1   */
	switch(endi & 0x00FF0000) {
	case 0x00010000:			/*  x	1   x	x   */
	    r |= (value >> 16) & 0x000000FF;
	    break;
	case 0x00020000:			/*  x	2   x	x   */
	    r |= (value >>  8) & 0x0000FF00;
	    break;
	case 0x00030000:			/*  x	3   x	x   */
	    r |= value & 0x00FF0000;
	    break;
	case 0x00040000:			/*  x	4   x	x   */
	    r |= (value <<  8) & 0xFF000000;
	    break;
	}
						/*  4	3   2	1   */
	switch(endi & 0x0000FF00) {
	case 0x00000100:			/*  x	x   1	x   */
	    r |= (value >>  8) & 0x000000FF;
	    break;
	case 0x00000200:			/*  x	x   2	x   */
	    r |= value & 0x0000FF00;
	    break;
	case 0x00000300:			/*  x	x   3	x   */
	    r |= (value <<  8) & 0x00FF0000;
	    break;
	case 0x00000400:			/*  x	x   4	x   */
	    r |= (value << 16) & 0xFF000000;
	    break;
	}
						/*  4	3   2	1   */
	switch(endi & 0x000000FF) {
	case 0x00000001:			/*  x	x   x	1   */
	    r |= value & 0x000000FF;
	    break;
	case 0x00000002:			/*  x	x   x	2   */
	    r |= (value <<  8) & 0x0000FF00;
	    break;
	case 0x00000003:			/*  x	x   x	3   */
	    r |= (value << 16) & 0x00FF0000;
	    break;
	case 0x00000004:			/*  x	x   x	4   */
	    r |= (value << 24) & 0xFF000000;
	    break;
	}
	break;
    }
    return(r);
}

#ifdef TEST

/*
 *  TEST the database
 *
 *  TEST l			list
 *  TEST a rid type "data"      append
 *  TEST m rid type "data"      modify
 *  TEST d rid			delete
 *  TEST c			clean
 *
 *  NOTE!  DO NOT SPECIFY ILLEGAL RID's!!!
 */

void ListTree(DBHan *, long, long);

DBHan *Db;

void
myexit()
{
    if (Db) {
	CloseDataBase(Db);
	Db = NULL;
    }
}

main(ac, av)
char *av[];
{
    record_t rid;

    if (ac == 1) {
	puts("TEST l");
	puts("TEST a rid type \"data\"");
	puts("TEST m rid type \"data\"");
	puts("TEST M par_rid rid newpar_rid");
	puts("TEST d rid");
	puts("TEST c");
	return(0);
    }
    atexit(myexit);

    if ((Db = OpenDataBase("t:test", "r")) == NULL)
	Db = OpenDataBase("t:test", "w");

    if (Db == NULL) {
	puts("Can't create t:test!");
	exit(1);
    }

    switch(av[1][0]) {
    case 'l':
	ListTree(Db, 0, 0);
	break;
    case 'a':
	rid = AppendRecord(Db, strtol(av[2], NULL, 0), strtol(av[3], NULL, 0), av[4], strlen(av[4]));
	break;
    case 'm':
	rid = ModifyRecord(Db, strtol(av[2], NULL, 0), strtol(av[3], NULL, 0), av[4], strlen(av[4]));
	break;
    case 'M':
	rid = MoveRecord(Db, strtol(av[2], NULL, 0), strtol(av[3], NULL, 0), strtol(av[4], NULL, 0));
	break;
    case 'd':
	DeleteRecord(Db, strtol(av[2], NULL, 0));
	break;
    case 'c':
	CloseDataBase(Db);
	Db = NULL;
	CleanDataBase("t:test");
	break;
    default:
	puts("bad command");
	exit(1);
    }
    return(0);
}

void
ListTree(db, rid, tab)
DBHan *db;
record_t rid;
long tab;
{
    char *ptr;
    long type;
    long size;

    for (rid = ScanRecord(db, rid, &size, &type); rid; rid = ScanNextRecord(db, rid, &size, &type)) {
	record_t subId;

	printf("%-6d%*.*s%c%08lx %-3d ", rid, tab, tab, "", (type & 0x80000000) ? 'D' : ' ', type, size);
	fflush(stdout);
	if (ptr = LoadRecord(db, rid, NULL, size)) {
	    printf("%.*s", size, ptr);
	    zfree(ptr);
	}
	puts("");

	if (subId = ScanSubRecord(db, rid, &size, &type))
	    ListTree(db, subId, tab + 1);
    }
}

#endif

