
/*
 *  DBCACHE.C
 *
 *  File Cache
 *
 *
 *  copen(), cclose(), cread(), cwrite(), cseek(), csize()
 */

#include "defs.h"

Prototype struct CHan *copen(const char *, int);
Prototype void cclose(struct CHan *);
Prototype long cread(struct CHan *, void *, long);
Prototype long cwrite(struct CHan *, void *, long);
Prototype long cseek(struct CHan *, long);
Prototype long csize(struct CHan *);
Prototype void dbc_atexit(void);

#define BLOCKBITS   10
#define BLOCKSIZE   (1 << BLOCKBITS)
#define BLOCKMASK   (BLOCKSIZE-1)
#define MAXBLOCKS   32
#define CACHESIZE   32
#define CACHEMASK   (CACHESIZE-1)

#define hash(offset)    (((offset) >> BLOCKBITS) & CACHEMASK)

typedef struct CBlock {
    struct CBlock *cb_Next;
    short   cb_Flags;
    long    cb_Offset;
    char    cb_Buf[BLOCKSIZE];
} CBlock;

#define CBF_DIRTY   0x0001

typedef struct CHan {
    Node    ch_Node;
    long    ch_Fh;	/*  file descriptor	    */
    long    ch_Size;	/*  size of file	    */
    long    ch_Seek;	/*  current seek position   */
    long    ch_FHSeek;	/*  real seek position in file	*/
    long    ch_FHSize;	/*  real file size		*/
    long    ch_NCBlks;	/*  # of cache blocks allocated */
    CBlock  *ch_Cache[CACHESIZE];
} CHan;

void FlushCCache(CHan *);
void ClearCCache(CHan *);
void CFill(long, long);
CBlock *GetCBlock(CHan *, long, long, long *, long *);

static List    DBCList;
static short   DBCAtExitIn;

void
dbc_atexit()
{
    CHan *cfh;

    if (DBCAtExitIn) {
	while (cfh = GetHead(&DBCList))
	    cclose(cfh);
    }
}

CHan *
copen(fileName, modes)
const char *fileName;
int modes;
{
    CHan *cfh = NULL;
    long fh;

    if (DBCAtExitIn == 0) {
	DBCAtExitIn = 1;
	NewList(&DBCList);
    }

    if (fh = Open(fileName, modes)) {
	cfh = zalloc(sizeof(CHan));
	cfh->ch_Fh = fh;
	Seek(fh, 0L, 1);
	cfh->ch_FHSize = cfh->ch_Size = Seek(fh, 0L, -1);
	AddTail(&DBCList, &cfh->ch_Node);
    }
    return(cfh);
}

void
cclose(cfh)
CHan *cfh;
{
    FlushCCache(cfh);
    ClearCCache(cfh);
    if (cfh->ch_Fh) {
	Close(cfh->ch_Fh);
	cfh->ch_Fh = 0;
    }
    Remove(&cfh->ch_Node);
    zfree(cfh);
}

long
cread(cfh, buf, bytes)
CHan *cfh;
void *buf;
long bytes;
{
    CBlock *cb;
    long blk_off;
    long blk_len;
    long n = 0;

    while (bytes && (cb = GetCBlock(cfh, cfh->ch_Seek, bytes, &blk_off, &blk_len))) {
	/*
	 *  check for read past end of file
	 */

	if (cfh->ch_Seek + blk_len > cfh->ch_Size) {
	    blk_len = cfh->ch_Size - cfh->ch_Seek;
	    if (blk_len <= 0)
		break;
	}

	/*
	 *  no, copy block
	 */

	movmem(cb->cb_Buf + blk_off, buf, blk_len);
	buf = (void *)((char *)buf + blk_len);
	bytes -= blk_len;
	cfh->ch_Seek += blk_len;
	n += blk_len;
    }
    return(n);
}

long
cwrite(cfh, buf, bytes)
CHan *cfh;
void *buf;
long bytes;
{
    CBlock *cb;
    long blk_off;
    long blk_len;
    long n = 0;

    while (bytes && (cb = GetCBlock(cfh, cfh->ch_Seek, bytes, &blk_off, &blk_len))) {
	cb->cb_Flags |= CBF_DIRTY;
	movmem(buf, cb->cb_Buf + blk_off, blk_len);
	buf = (void *)((char *)buf + blk_len);
	cfh->ch_Seek += blk_len;
	n += blk_len;
	bytes -= blk_len;

	/*
	 *  If extending the file must be sure ch_Size reflects this.
	 */

	if (cfh->ch_Seek > cfh->ch_Size)
	    cfh->ch_Size = cfh->ch_Seek;
    }
    return(n);
}

long
cseek(cfh, pos)
CHan *cfh;
long pos;
{
    cfh->ch_Seek = pos;
}

long
csize(cfh)
CHan *cfh;
{
    return(cfh->ch_Size);
}

void
FlushCCache(cfh)
CHan *cfh;
{
    CBlock *cb;
    short i;

    for (i = 0; i < CACHESIZE; ++i) {
	for (cb = cfh->ch_Cache[i]; cb; cb = cb->cb_Next) {
	    if (cb->cb_Flags & CBF_DIRTY) {
		/*
		 *  Write out the dirty block, do not write past our
		 *  virtual end of a file
		 */

		long bytes = BLOCKSIZE;

		cb->cb_Flags &= ~CBF_DIRTY;

		if (cb->cb_Offset + bytes > cfh->ch_Size)
		    bytes = cfh->ch_Size - cb->cb_Offset;
		if (bytes > 0) {
		    if (cfh->ch_FHSeek != cb->cb_Offset) {
			if (cb->cb_Offset > cfh->ch_FHSize) {
			    Seek(cfh->ch_Fh, cfh->ch_FHSize, -1);
			    CFill(cfh->ch_Fh, cb->cb_Offset - cfh->ch_FHSize);
			} else {
			    Seek(cfh->ch_Fh, cb->cb_Offset, -1);
			}
		    }
		    Write(cfh->ch_Fh, cb->cb_Buf, bytes);
		    cfh->ch_FHSeek = cb->cb_Offset + bytes;
		    if (cfh->ch_FHSeek > cfh->ch_FHSize)
			cfh->ch_FHSize = cfh->ch_FHSeek;
		}
	    }
	}
    }
}

void
ClearCCache(cfh)
CHan *cfh;
{
    CBlock *cb;
    short i;

    for (i = 0; i < CACHESIZE; ++i) {
	while (cb = cfh->ch_Cache[i]) {
	    cfh->ch_Cache[i] = cb->cb_Next;
	    zfree(cb);
	    --cfh->ch_NCBlks;
	}
    }
}

/*
 *  Core of the caching system
 *
 *  Given pos, *pbytes on arbitrary offsets find the first cache block
 *  that overlaps the specified range.
 */


CBlock *
GetCBlock(cfh, pos, bytes, pblk_off, pblk_len)
CHan *cfh;
long pos;
long bytes;
long *pblk_off;
long *pblk_len;
{
    long offset = pos & ~BLOCKMASK;
    CBlock *cb;
    CBlock **cbp;

    *pblk_off = pos - offset;	    /*	offset into block   */

    if (*pblk_off + bytes > BLOCKSIZE)
	*pblk_len = BLOCKSIZE - *pblk_off;
    else
	*pblk_len = bytes;

    for (cbp = cfh->ch_Cache + hash(offset); cb = *cbp; cbp = &cb->cb_Next) {
	if (cb->cb_Offset == offset)
	    break;
    }

    /*
     *	If cb == NULL then we have to create a cache block
     */

    if (cb == NULL) {
	long cbytes;

	if (cfh->ch_NCBlks >= MAXBLOCKS) {
	    FlushCCache(cfh);
	    ClearCCache(cfh);
	    cbp = cfh->ch_Cache + hash(offset);
	}
	*cbp = cb = zalloc(sizeof(CBlock));
	cb->cb_Offset = offset;
	++cfh->ch_NCBlks;

	cbytes = BLOCKSIZE;
	if (offset + cbytes > cfh->ch_FHSize)
	    cbytes = cfh->ch_FHSize - offset;
	if (cbytes > 0) {
	    if (cfh->ch_FHSeek != offset)
		Seek(cfh->ch_Fh, cb->cb_Offset, -1);
	    Read(cfh->ch_Fh, cb->cb_Buf, cbytes);
	    cfh->ch_FHSeek = cb->cb_Offset + cbytes;
	}
    }
    return(cb);
}

void
CFill(fh, bytes)
long fh;
long bytes;
{
    char *buf = zalloc(512);
    while (bytes) {
	long n = (bytes > 512) ? 512 : bytes;
	Write(fh, buf, n);
	bytes -= n;
    }
    zfree(buf);
}

