
/*
 *  FILES.C
 *
 *  File support routines
 */

#include "defs.h"

Prototype void GetFileInfo(RegNode *, CtlNode *, FileNode *, List *);
Prototype FileNode *MakeFindRecFileNode(CtlNode *, char *, short);
Prototype void IncorporateRecFileNode(CtlNode *, FileNode *, FileChunk *, void *);
Prototype void CalculateBytesToSend(RegNode *, CtlNode *);
Prototype void *SelectChunkForSend(RegNode *, CtlNode *, FileNode **, FileChunk **);
Prototype FILE *CreateFilePath(char *, long);
Prototype void CreateDirPath(char *);
Prototype long FileBytesRemaining(FileNode *);
Prototype void FileNodeChunksCompleted(FileNode *, short);

Prototype bool_t ValidFile(RegNode *, const char *);
Prototype void AssignFileNumbers(List *);

/*
 *  Get information on a file including whether it is valid for the
 *  purposes of a remote request
 *
 *  if list != NULL and the file is a directory, scan the directory
 *  creating additional FileNode structures.
 */

void
GetFileInfo(rn, cn, fn, list)
RegNode *rn;
CtlNode *cn;
FileNode *fn;
List *list;
{
    BPTR lock;
    __aligned FileInfoBlock fib;

    fn->fn_Flags = FNF_ILLEGAL | FNF_BAD | GF_MODIFIED;

    if (ValidFile(rn, fn->fn_LocFileName) == 0) {
	fn->fn_Flags &= ~FNF_ILLEGAL;
	if (lock = Lock(fn->fn_LocFileName, SHARED_LOCK)) {
	    if (Examine(lock, &fib)) {
		fn->fn_Flags &= ~FNF_BAD;
		fn->fn_Flags |= FNF_SIZE;
		if (fib.fib_DirEntryType > 0)
		    fn->fn_Flags |= FNF_DIR;
		fn->fn_Protection = fib.fib_Protection;
		movmem(&fib.fib_Date, fn->fn_Date, sizeof(long) * 3);
		fn->fn_Size = fib.fib_Size;

		if (fib.fib_DirEntryType > 0 && list) {
		    while (ExNext(lock, &fib)) {
			FileNode *fn2 = AllocFileNode(cn);
			short len = strlen(fn->fn_LocFileName);

			fn2->fn_LocFileName = zalloc(len + strlen(fib.fib_FileName) + 2);
			strcpy(fn2->fn_LocFileName, fn->fn_LocFileName);
			if (len && fn->fn_LocFileName[len-1] != ':' && fn->fn_LocFileName[len-1] != '/')
			    strcat(fn2->fn_LocFileName, "/");
			strcat(fn2->fn_LocFileName, fib.fib_FileName);

			GetFileInfo(rn, cn, fn2, list);
			if (fn2->fn_Flags & (FNF_ILLEGAL|FNF_BAD))
			    UnLoadBMSFileTree(fn2);
		    }
		}
	    }
	    UnLock(lock);
	}
    }
}

/*
 *  Find file node at Receiver.  The remote's filepath is used since that
 *  is all the remote knows.
 */

FileNode *
MakeFindRecFileNode(cn, fileName, isDir)
CtlNode *cn;
char *fileName;
short isDir;
{
    FileNode *fn;

    for (fn = GetHead(&cn->cn_FileList); fn; fn = GetSucc(&fn->fn_Node)) {
	if (strcmp(fileName, fn->fn_RemFileName) == 0)
	    break;
    }
    if (fn == NULL) {
	FileNode *fnTmp;

	fn = AllocFileNode(cn);
	fn->fn_RemFileName = zstrdup(fileName);

	/*
	 *  generate the local file name.  e.g.
	 *	    MASTER:    local=ram:xx remote=fubar:a/b/c
	 *	    fileName:			   fubar:a/b/c/d
	 *	    reuslt:    ram:xx/d
	 */

	for (fnTmp = GetHead(&cn->cn_FileList); fnTmp; fnTmp = GetSucc(&fnTmp->fn_Node)) {
	    if ((fnTmp->fn_Flags & FNF_MASTER) && strnicmp(fnTmp->fn_RemFileName, fileName, strlen(fnTmp->fn_RemFileName)) == 0)
		break;
	}
	if (fnTmp) {
	    fn->fn_LocFileName = zalloc(strlen(fnTmp->fn_LocFileName) + strlen(fileName) - strlen(fnTmp->fn_RemFileName) + 8);
	    strcpy(fn->fn_LocFileName, fnTmp->fn_LocFileName);
	    strcat(fn->fn_LocFileName, fileName + strlen(fnTmp->fn_RemFileName));

	    /*
	     *	munge local file name until reception is complete
	     */

	    if ((fn->fn_Flags & FNF_DONE) == 0 && isDir == 0)
		strcat(fn->fn_LocFileName, "-XXX");
	} else {
	    bmslog("GenFile", "Received file that does not match request: %s", fileName);
	    fn->fn_LocFileName = zstrdup("bms:file.bad");
	}
    }
    fn->fn_Flags |= GF_MODIFIED;
    return(fn);
}

/*
 *  Incorporate file/dir info and data as appropriate.	ffc/data will
 *  be NULL for directories.
 */

void
IncorporateRecFileNode(cn, fn, ffc, data)
CtlNode *cn;
FileNode *fn;
FileChunk *ffc;
void *data;
{
    FileChunk *fc;
    FILE *fi;

    fn->fn_Flags |= GF_MODIFIED;

    if (ffc) {
	/*
	 *  files are made write-only until complete.
	 */

	if ((fn->fn_Flags & (FNF_ILLEGAL|FNF_BAD)) == 0) {
	    if (fi = CreateFilePath(fn->fn_LocFileName, ffc->fc_Offset + ffc->fc_Bytes)) {
		fseek(fi, ffc->fc_Offset, 0);
		fwrite(data, ffc->fc_Bytes, 1, fi);
		fclose(fi);
		fc = AllocFileChunk(fn, ffc->fc_Offset, ffc->fc_Bytes);
		fc->fc_Flags |= ffc->fc_Flags | GF_MODIFIED;
	    }
	}
	if (FileBytesRemaining(fn) == 0)
	    FileNodeChunksCompleted(fn, 1);
    } else {
	if ((fn->fn_Flags & (FNF_ILLEGAL|FNF_BAD)) == 0)
	    CreateDirPath(fn->fn_LocFileName);
	fn->fn_Flags |= FNF_DONE | GF_MODIFIED;
    }
    fn->fn_Flags |= GF_MODIFIED;
}

/*
 *  For transmit side.	If 0 then CNF_DONE set
 */

void
CalculateBytesToSend(rn, cn)
RegNode *rn;
CtlNode *cn;
{
    long n = 0;
    FileNode *fn;

    for (fn = GetHead(&cn->cn_FileList); fn; fn = GetSucc(&fn->fn_Node)) {
	n += FileBytesRemaining(fn);

	/*
	 *  A zero here dictates whether we are done.  Therefore, ensure
	 *  that we do not get zero if we aren't done!  Directories and
	 *  bad files have zero bytes but must still be processed
	 */

	if ((fn->fn_Flags & FNF_DONE) == 0 && (fn->fn_Flags & (FNF_DIR|FNF_BAD|FNF_ILLEGAL)))
	    ++n;
    }
    cn->cn_Comm.com_TxReq = n;
    if (n == 0)
	cn->cn_Flags |= CNF_DONE;
    cn->cn_Flags |= GF_MODIFIED;
    cn->cn_Comm.com_Flags |= GF_MODIFIED;
}

/*
 *  Select a chunk of data to send
 */

void *
SelectChunkForSend(rn, cn, pfn, pfc)
RegNode *rn;
CtlNode *cn;
FileNode **pfn;
FileChunk **pfc;
{
    FileNode *fn;
    FileChunk *fc;

    if (DebugOpt)
	bmslog("SelectChunk", "%d bytes available", cn->cn_Comm.com_TxOk);

    if (cn->cn_Comm.com_TxOk == 0)
	return(NULL);

    cn->cn_Comm.com_Flags |= GF_MODIFIED;

    for (fn = GetHead(&cn->cn_FileList); fn; fn = GetSucc(&fn->fn_Node)) {
	long offset = 0;
	long bytes = 0;

	if (DebugOpt)
	    printf("FN %s %04x\n", fn->fn_LocFileName, fn->fn_Flags);

	if (fn->fn_Flags & FNF_DONE)
	    continue;

	/*
	 *  If a directory, illegal, or bad then there is no data
	 *  associated with the file, set done flag and return now.
	 */

	if (fn->fn_Flags & (FNF_DIR|FNF_ILLEGAL|FNF_BAD)) {
	    fn->fn_Flags |= FNF_DONE | GF_MODIFIED;
	    *pfn = fn;
	    *pfc = NULL;
	    return(zalloc(1));
	}

	for (fc = GetHead(&fn->fn_ChunkList); fc; fc = GetSucc(&fc->fc_Node)) {
	    if (fc->fc_Flags & GF_DELETED)
		continue;
	    if (offset < fc->fc_Offset) {
		bytes = fc->fc_Offset - offset;

		/*
		 *  Check bytes verses number of bytes we are allowed to
		 *  send
		 */

		if (bytes > cn->cn_Comm.com_TxOk)
		    bytes = cn->cn_Comm.com_TxOk;
		break;
	    }
	    offset = fc->fc_Offset + fc->fc_Bytes;
	}

	/*
	 *  If file is all sent
	 *
	 */

	if (offset == fn->fn_Size) {
	    fn->fn_Flags |= FNF_DONE | GF_MODIFIED;
	    /*FileNodeChunksCompleted(fn, 0);*/
	    continue;
	}

	if (MailBytes <= MinMailBytes)      /*  not enough mail space   */
	    continue;

	if (bytes == 0) {                   /*  no chunks (yet)         */
	    bytes = fn->fn_Size - offset;
	    if (bytes > cn->cn_Comm.com_TxOk)
		bytes = cn->cn_Comm.com_TxOk;
	}
	if (bytes > cn->cn_Comm.com_TxOk) { /*  how many can we send?   */
	    if (cn->cn_Comm.com_TxOk < MinMailBytes) {
		if (bytes > MinMailBytes)
		    bytes = MinMailBytes;
	    } else {
		bytes = cn->cn_Comm.com_TxOk;
	    }
	}
	if (bytes > MailBytes) {            /*  mail file maximum?      */
	    if (bytes > MinMailBytes) {
		if (MailBytes < MinMailBytes)
		    bytes = MinMailBytes;
		else
		    bytes = MailBytes;
	    }
	}
	MailBytes -= bytes;

	if (bytes == 0)
	    continue;

	cn->cn_Comm.com_TxOk -= bytes;

	/*
	 *  Can send <bytes> bytes of data from offset <offset>
	 *  XXX what if file does not exist?
	 */

	{
	    FILE *fi = fopen(fn->fn_LocFileName, "r+");
	    char *ptr;

	    fseek(fi, offset, 0);
	    ptr = zalloc(bytes);
	    fread(ptr, bytes, 1, fi);
	    fclose(fi);

	    *pfn = fn;
	    *pfc = fc = AllocFileChunk(fn, offset, bytes);
	    return(ptr);
	}
    }
    return(NULL);
}

FILE *
CreateFilePath(fileName, fileSize)
char *fileName;
long fileSize;
{
    FILE *fi;
    char *ptr;

    if ((fi = fopen(fileName, "r+")) == NULL) {
	if ((fi = fopen(fileName, "w+")) == NULL) {
	    if (ptr = strrchr(fileName, '/')) {
		*ptr = 0;
		CreateDirPath(fileName);
		*ptr = '/';
	    }
	    fi = fopen(fileName, "w+");
	}
    }
    if (fi) {
	long n;

	fseek(fi, 0L, 2);
	n = ftell(fi);
	if (n < fileSize) {
	    clrmem(TmpBuf2, sizeof(TmpBuf2));

	    while (n < fileSize) {
		long r = fileSize - n;
		if (r > sizeof(TmpBuf2))
		    r = sizeof(TmpBuf2);
		fwrite(TmpBuf2, r, 1, fi);
		n += r;
	    }
	}
    } else {
	bmslog("CreateFile", "Unable to create file: %s", fileName);
    }
    return(fi);
}

void
CreateDirPath(fileName)
char *fileName;
{
    BPTR lock;
    char *ptr;
    short i;

    if (lock = Lock(fileName, SHARED_LOCK)) {
	UnLock(lock);
	return;
    }
    if (ptr = strrchr(fileName, '/')) {
	*ptr = 0;
	CreateDirPath(fileName);
	*ptr = '/';
    }
    if (lock = CreateDir(fileName)) {
	UnLock(lock);
	return;
    }
    bmslog("CreateDir", "Unable to create dir: %s", fileName);
}

/*
 *  Return the number of bytes remaining, -1 if unknown.  The number
 *  of bytes remaining is always known for files we send, may be
 *  unknown for files we receive.
 */

long
FileBytesRemaining(fn)
FileNode *fn;
{
    FileChunk *fc;
    long bytes = 0;
    long offset = 0;

    if (DebugOpt)
	printf("FileBytesRemaining flags=%04x %s\n", fn->fn_Flags, fn->fn_LocFileName);
    if (fn->fn_Flags & (FNF_DONE|FNF_ILLEGAL|FNF_BAD))
	return(0);
    if ((fn->fn_Flags & FNF_SIZE) == 0)
	return(-1);

    /*
     *	Loop must take into account overlapping chunks
     */

    for (fc = GetHead(&fn->fn_ChunkList); fc; fc = GetSucc(&fc->fc_Node)) {
	if (fc->fc_Flags & GF_DELETED)
	    continue;
	if (offset < fc->fc_Offset + fc->fc_Bytes) {
	    if (offset < fc->fc_Offset)
		bytes -= fc->fc_Offset - offset;
	    bytes += fc->fc_Offset + fc->fc_Bytes - offset;
	    offset = fc->fc_Offset + fc->fc_Bytes;
	}
    }
    if (DebugOpt)
	printf("Bytes remaining: %d/%d (%d left)\n", bytes, fn->fn_Size, fn->fn_Size - bytes);
    return(fn->fn_Size - bytes);
}

void
FileNodeChunksCompleted(fn, mungeName)
FileNode *fn;
short mungeName;
{
    FileChunk *fc;

    if (mungeName && (fn->fn_Flags & (FNF_DONE|FNF_DIR)) == 0) {
	/*
	 *  'fix' local file name
	 */
	short len = strlen(fn->fn_LocFileName);
	if (len > 4 && strcmp(fn->fn_LocFileName + len - 4, "-XXX") == 0) {
	    char *ptr = zstrdup(fn->fn_LocFileName);
	    ptr[len-4] = 0;
	    remove(ptr);
	    rename(fn->fn_LocFileName, ptr);
	    fn->fn_Flags |= GF_MODIFIED;
	    zfree(fn->fn_LocFileName);
	    fn->fn_LocFileName = ptr;
	}
    }
    fn->fn_Flags |= FNF_DONE | GF_MODIFIED;
    for (fc = GetHead(&fn->fn_ChunkList); fc; fc = GetSucc(&fc->fc_Node))
	fc->fc_Flags |= GF_DELETED;
}

/*
 *  Check to see if a file is valid, return -1 on failure, 0 on success
 */

bool_t
ValidFile(rn, fileName)
RegNode *rn;
const char *fileName;
{
    short maxlen = 0;

    {
	short len;
	FileAccess *fa;
	FileAccess *fgot = NULL;

	for (fa = GetHead(&rn->rn_AccList); fa; fa = GetSucc(&fa->fa_Node)) {
	    if (fa->fa_Flags & GF_DELETED)
		continue;
	    len = strlen(fa->fa_Node.ln_Name);
	    if (len >= maxlen && strnicmp(fa->fa_Node.ln_Name, fileName, len) == 0) {
		maxlen = len;
		fgot = fa;
	    }
	}
	if (fgot == NULL) {
	    if (rn != DefRN)
		return(ValidFile(DefRN, fileName));
	    return(-1);
	}
	if (fgot->fa_Flags & FAF_RESTRICT)
	    return(-1);
    }

    /*
     *	Make sure that remainder of file does not contain invalid characters
     *	(something like fubar// to get back a directory or a colon)
     */

    {
	char *ptr;

	for (ptr = fileName + maxlen; *ptr; ++ptr) {
	    switch(*ptr) {
	    case ':':
		return(-1);
	    case '/':
		if (ptr[1] == 0 || ptr[1] == '/')
		    return(-1);
		break;
	    }
	}
    }
    return(0);
}

/*
 *  Assign file numbers
 */

void
AssignFileNumbers(list)
List *list;
{
    FileNode *fn;
    long numFiles = 0;
    long nfno = 0;

    for (fn = GetHead(list); fn; fn = GetSucc(&fn->fn_Node))
	++numFiles;
    for (fn = GetHead(list); fn; fn = GetSucc(&fn->fn_Node)) {
	fn->fn_Files = numFiles;
	fn->fn_FileNo = nfno;
	fn->fn_Flags |= GF_MODIFIED;
	++nfno;
    }
}

