
/*
 *  BATCHNEWS.C
 *
 *  (C)Copyright 1990, Mark R. Rinfret, All Rights Reserved
 *
 *  Author:	Mark R. Rinfret, September 1990
 *  library compress routines melded in Apr 1991, Matt Dillon
 *  Modified May 1991 by Henning Schmiedehausen
 *
 *  The batchnews program allows you to forward news to other sites.
 *  It expects to find a text file named UUSPOOL/batch/<system>
 *  where <system> is the name of the system to receive the news.
 *  The file UULIB/sys is the configuration file which determines the
 *  newsgroups that are to be forwarded to <system>.
 */

#include "news.h"
#include <errno.h>
#include <ctype.h>
#include <log.h>
#include <exec/types.h>
#include <libraries/dos.h>
#include <owndevunit.h>

IDENT(".09");

void		BatchNews(const char *dirName, const char *systemName);
void		SpoolBatch(const char *systemName);

typedef struct FileInfoBlock	FIB;

static char TmpBuf[2048];
static char TmpBuf2[256];
char homedir[256];

char *NewsDir;
char *WKName;
struct Library *OwnDevUnitBase;

#define PROG_NAME "BatchNews"

int _bufsiz = 8192;
short	Bits;
long	SpoolSize;

void myexit(void);

int
main(ac, av)
int ac;
char *av[];
{
    char	*batchDirName;
    FIB 	*fib = malloc(sizeof(FIB));
    int 	ret = 0;
    short	i;
    BPTR	lock;

    LogProgram = PROG_NAME;

    atexit(myexit);
    if ((OwnDevUnitBase = OpenLibrary(ODU_NAME, 0)) == NULL) {
	puts("Couldn't open lock library");
	exit(20);
    }

    LockFile("BATCHNEWS");
#ifndef TEST
    fclose(stderr);         /* Assume we are running in the background */
#endif

    getcwd(homedir, sizeof(homedir));
    NewsDir = GetConfigDir(UUNEWS);
    Bits = atoi(GetConfig(BATCHBITS, "12"));
    SpoolSize = atoi(GetConfig(SPOOLSIZE, "65000"));

    if (Bits < 12 || Bits > 16)
	Bits = 12;
    if (SpoolSize <= 100)
	SpoolSize = 16000;

    if (chdir(NewsDir) != 0) {
	ulog(-1, "BatchNews: NewsDir '%s' doesn't exist; quitting", NewsDir);
	exit(1);
    }

    WKName = strdup(TmpFileName("news-work"));

    batchDirName = strdup(MakeConfigPath(UUSPOOL,"batch"));
    lock = Lock(batchDirName, SHARED_LOCK);
    if (! lock ) {
	ulog(-1,"Where is batch directory '%s'?", batchDirName);
	exit(1);
    }
    if (Examine(lock, fib)) {
	while (ExNext(lock, fib)) {
	    if (fib->fib_DirEntryType > 0)
		continue;
#ifdef TEST
	    printf("scan: '%s'\n", fib->fib_FileName);
#endif
	    if (is_in_L_sys_file(fib->fib_FileName)) {
		BatchNews(batchDirName, fib->fib_FileName);
	    }
	}
    }
    UnLock(lock);
    exit(ret);
}

void
myexit()
{
    chdir(homedir);

    UnLockFiles();

    if (OwnDevUnitBase) {
	CloseLibrary(OwnDevUnitBase);
	OwnDevUnitBase = NULL;
    }
}

void
BatchNews(const char *dirName, const char *systemName)
{

    static char batchFileName[128];

    char *nodename = FindConfig(NODENAME);
    char *domainname= FindConfig(DOMAINNAME);
    char *domainpath = GetConfig(DOMAINPATH, "N");
    FILE	    *batchFile = NULL;
    FIB 	    *fib = malloc(sizeof(FIB));
    long	    fileSize;
    int 	    length;
    BPTR	    lock = 0;
    FILE	    *newsFile = NULL;
    int 	    ret = 0;
    long	    totalSize;
    FILE	    *workFile = NULL;
    long	    workPos;
    char	    *newsFileName;

    if (domainpath[0] == '1' || domainpath[0] == 'y' || domainpath[0] == 'Y')
	domainpath = domainname;
    else
	domainpath = "";

#ifdef TEST
    printf("BatchNews: system = '%s'\n", systemName);
#endif
    strcpy(batchFileName, dirName);
    strcat(batchFileName, "/");
    strcat(batchFileName, systemName);
    batchFile = fopen(batchFileName, "r");
    if (! batchFile) {
#ifdef TEST
	perror(batchFileName);
#endif
	return; 	   /* Try later? */
    }

    workFile = fopen(WKName, "w");
    if (! workFile) {
no_workfile:
	ret = errno;
	ulog(-1,"Failed to open work file '%s'\n",WKName);
#ifdef TEST
	perror(WKName);
#endif
	exit(ret);
    }

    totalSize = 0;

    while (fgets(TmpBuf, 80, batchFile)) {
#ifdef TEST
	puts(TmpBuf);
#endif
	if (*TmpBuf == '\0' || *TmpBuf == '\n')
	    continue;

	length = strlen(TmpBuf);

	if (TmpBuf[length-1] == '\n')
	    TmpBuf[--length] = '\0';        /* Drop trailing newline. */

	if (newsFileName = HandleHeirarchy(GetConfigDir(UUNEWS), TmpBuf, 1)) {
	    strcpy(TmpBuf, newsFileName);
	    free(newsFileName);

	    lock = Lock(TmpBuf, SHARED_LOCK);
	    if (!lock) {
bad_article:
		ulog(-1,"Article '%s' missing?", TmpBuf);
#ifdef TEST
		printf("Bad article name: '%s'\n", TmpBuf);
#endif
		continue;
	    }
	    if (Examine(lock, fib)) {
		fileSize = fib->fib_Size;
	    } else {
		UnLock(lock);
		goto bad_article;
	    }
	    UnLock(lock);

	    if (fileSize + totalSize + 64 > SpoolSize) {
		fclose(workFile);
		SpoolBatch(systemName);
		totalSize = 0;
		workFile = fopen(WKName, "w");  /* Re-open the work file. */
		if (!workFile)
		    goto no_workfile;
	    }
	    newsFile = fopen(TmpBuf, "r");
	}

	/*
	 *  newsFileName is invalid after this except for boolean checks
	 */

	if (newsFileName == NULL || newsFile == NULL)
	    continue;

	/*
	 *  Handle news file.  Delete Xref: lines, modify Path: lines,
	 *  copy rest.	Filesize is modified at seekback.
	 */

	workPos = ftell(workFile);
	fprintf(workFile,"#! rnews %06ld\n", 0);     /*  rewritten later */
	fileSize = 0;

	while (fgets(TmpBuf, sizeof(TmpBuf) - 16, newsFile)) {
	    if (strnicmp(TmpBuf, "Xref:", 5) == 0)
		continue;
	    if (strncmp(TmpBuf, "Path:", 5) == 0) {
		char *ptr = TmpBuf + 5;
		while (*ptr == ' ' || *ptr == 9)
		    ++ptr;
		strins(ptr, "!");
		strins(ptr, domainpath);
		strins(ptr, nodename);
	    }
	    fputs(TmpBuf, workFile);
	    fileSize += strlen(TmpBuf);
	    if (TmpBuf[0] == '\n')
		break;
	}
	while (fgets(TmpBuf, sizeof(TmpBuf) - 16, newsFile)) {
	    fputs(TmpBuf, workFile);
	    fileSize += strlen(TmpBuf);
	}
	fclose(newsFile);
	totalSize += fileSize;
	fseek(workFile, workPos, 0);
	fprintf(workFile,"#! rnews %06ld\n", fileSize);
	fseek(workFile, 0, 2);
    }

done:
    if (batchFile) {
	fclose(batchFile);
	remove(batchFileName);
    }
    if (workFile) {
	fclose(workFile);
	if (ret == 0 && totalSize)
	    SpoolBatch(systemName);
    }
    free(fib);
    remove(WKName);
}

/*
 *  modified 1.14 to use library
 *
 */

void
SpoolBatch(const char *systemName)
{
    FILE *fo;
    int r = -1;

    sprintf(TmpBuf2, "%s.Z", WKName);
    if (fo = fopen(TmpBuf2, "w")) {
	fputs("#! cunbatch\n", fo);
	r = compress_from_file(WKName, fo, Bits);
	fclose(fo);
    }
    if (r == 0) {
	remove(WKName);
	rename(TmpBuf2, WKName);
    } else {
	remove(TmpBuf2);
	ulog(-1, "compress failed, low memory or uncompressable file");
	ulog(-1, "will proceed with an uncompressed batch");
    }

    /* Submit it to uux for transmission to target system. */
    sprintf(TmpBuf2, "%s %s \"%s!rnews\"",
	GetConfigProgram(UUX),
	WKName,
	systemName
    );
    system(TmpBuf2);
}

